#!/usr/bin/env python
#
# Copyright (c) 2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
This file contains the replicate utility. It is used to establish a
multi-source replication topology.
"""

from mysql.utilities.common.tools import check_python_version

# Check Python version compatibility
check_python_version()

import os.path
import sys
import logging

from mysql.utilities.exception import FormatError, UtilError, UtilRplError
from mysql.utilities.common.ip_parser import parse_connection
from mysql.utilities.common.messages import (
    PARSE_ERR_OPTS_REQ_GREATER_OR_EQUAL,
    PARSE_ERR_OPTS_REQ
)
from mysql.utilities.common.options import setup_common_options
from mysql.utilities.common.options import (add_verbosity, add_rpl_user,
                                            add_format_option)
from mysql.utilities.common.server import check_hostname_alias
from mysql.utilities.common.tools import check_connector_python
from mysql.utilities.common.my_print_defaults import MyDefaultsReader
from mysql.utilities.command.rpl_admin import purge_log
from mysql.utilities.command.setup_rpl import start_ms_replication


# Constants
NAME = "MySQL Utilities - mysqlrplms "
DESCRIPTION = "mysqlrplms - establish multi-source replication"
USAGE = ("%prog --slave=root@localhost:3306 --masters=root@localhost:3310,"
         "root@localhost:3311 --rpl-user=rpl:passwd")
DATE_FORMAT = '%Y-%m-%d %H:%M:%S %p'
EXTENDED_HELP = """
Introduction
------------
The mysqlrplms utility is used to setup round robin multi-source replcation.
This technique can be a solution for aggregating streams of data from multiple
masters for a single slave.

The mysqlrplms utility follows these assumptions:

  - All servers have GTIDs enabled.
  - There are no conflicts between transactions from different sources/masters.
    For example, there are no updates to the same object from multiple masters.
  - Replication is asynchronous.

A round-robin scheduling is used to setup replication among the masters and
slave.

The utility can be run as a daemon on POSIX systems.

  # Basic multi-source replication setup.

  $ mysqlrplms --slave=root:pass@host1:3306 \\
               --masters=root:pass@host2:3306,root:pass@host3:3306

  # Multi-source replication setup using a different report values.

  $ mysqlrplms --slave=root:pass@host1:3306 \\
               --masters=root:pass@host2:3306,root:pass@host3:3306 \\
               --report-values=health,gtid,uuid

  # Start multi-source replication running as a daemon. (POSIX only)

  $ mysqlrplms --slave=root:pass@host1:3306 \\
               --masters=root:pass@host2:3306,root:pass@host3:3306 \\
               --log=rplms_daemon.log --pidfile=rplms_daemon.pid \\
               --daemon=start

  # Restart a multi-source replication running as a daemon.

  $ mysqlrplms --slave=root:pass@host1:3306 \\
               --masters=root:pass@host2:3306,root:pass@host3:3306 \\
               --log=rplms_daemon.log --pidfile=rplms_daemon.pid \\
               --daemon=restart

  # Stop a multi-source replication running as a daemon.

  $ mysqlrplms --slave=root:pass@host1:3306 \\
               --masters=root:pass@host2:3306,root:pass@host3:3306 \\
               --log=rplms_daemon.log --pidfile=rplms_daemon.pid \\
               --daemon=stop


Helpful Hints
-------------
  - The default report value is 'health'.
    This value can be changed with the --report-values option. It can be
    'health', 'gtid' or 'uuid'. Multiple values can be used separated by
    commas.

  - The default output for reporting health is 'grid'.
    This value can be changed with the --format option. It can be 'grid',
    'tab', 'csv' or 'vertical' format.

  - The default interval for reporting health is 15 seconds.
    This value can be changed with the --interval option.

  - The default interval for switching masters is 60 seconds.
    This value can be changed with the --switchover-interval option.

"""

# Check for connector/python
if not check_connector_python():
    sys.exit(1)

if __name__ == '__main__':
    # Setup the command parser
    parser = setup_common_options(os.path.basename(sys.argv[0]),
                                  DESCRIPTION, USAGE, True, False,
                                  extended_help=EXTENDED_HELP)

    # Setup utility-specific options:

    # Interval for reporting health
    parser.add_option("--interval", "-i", action="store", dest="interval",
                      type="int", default="15", help="interval in seconds for "
                      "reporting health. Default = 15 seconds. "
                      "Lowest value is 5 seconds.")

    # Interval for switching masters
    parser.add_option("--switchover-interval", action="store",
                      dest="switchover_interval",
                      type="int", default="60", help="interval in seconds for "
                      "switching masters. Default = 60 seconds. "
                      "Lowest value is 30 seconds.")

    # Connection information for the sink server
    parser.add_option("--slave", action="store", dest="slave",
                      type="string", default=None,
                      help="connection information for slave server in "
                      "the form: <user>[:<password>]@<host>[:<port>]"
                      "[:<socket>] or <login-path>[:<port>][:<socket>].")

    # Connection information for the masters servers
    parser.add_option("--masters", action="store", dest="masters",
                      type="string", default=None, help="connection "
                      "information for master servers in the form: "
                      "<user>[:<password>]@<host>[:<port>][:<socket>] or "
                      "<login-path>[:<port>][:<socket>]. List multiple master "
                      "in comma-separated list.")

    # Replication user and password
    add_rpl_user(parser)

    # Add start from beginning option
    parser.add_option("-b", "--start-from-beginning", action="store_true",
                      default=False, dest="from_beginning",
                      help="start replication from the first event recorded "
                      "in the binary logging of the masters.")

    # Add report values
    parser.add_option("--report-values", action="store", dest="report_values",
                      type="string", default="health",
                      help="report values used in multi-source replication. "
                      "It can be health, gtid or uuid. Multiple values can be "
                      "used separated by commas. The default is health.")

    # Add output format
    add_format_option(parser, "display the output in either grid (default), "
                      "tab, csv, or vertical format", None)

    # Add option to run as daemon
    parser.add_option("--daemon", action="store", dest="daemon", default=None,
                      help="run on daemon mode. It can be start, stop, "
                      "restart or nodetach.", type="choice",
                      choices=("start", "stop", "restart", "nodetach"))

    # Add pidfile for the daemon option
    parser.add_option("--pidfile", action="store", dest="pidfile",
                      type="string", default=None, help="pidfile for running "
                      "mysqlrplms as a daemon.")

    # Add a log file to use for logging messages
    parser.add_option("--log", action="store", dest="log_file", default=None,
                      type="string", help="specify a log file to use for "
                      "logging messages")

    # Add the maximum age of log entries in days for the logging system
    parser.add_option("--log-age", action="store", dest="log_age", default=7,
                      type="int", help="specify maximum age of log entries in "
                      "days. Entries older than this will be purged on "
                      "startup. Default = 7 days.")

    # Add verbosity
    add_verbosity(parser)

    # Now we process the rest of the arguments.
    opt, args = parser.parse_args()

    # Check if the values specified for the --report-values option are valid.
    for report in opt.report_values.split(","):
        if report.lower() not in ("health", "gtid", "uuid"):
            parser.error("The value for the option --report-values is not "
                         "valid: '{0}', the values allowed are 'health', "
                         "'gitd' or 'uuid'".format(opt.report_values))

    # Check for errors
    if int(opt.interval) < 5:
        parser.error(PARSE_ERR_OPTS_REQ_GREATER_OR_EQUAL.format(
            opt="--interval", value=5))

    if int(opt.switchover_interval) < 30:
        parser.error(PARSE_ERR_OPTS_REQ_GREATER_OR_EQUAL.format(
            opt="--switchover-interval", value=30))

    # option --slave is required (mandatory)
    if not opt.slave:
        parser.error(PARSE_ERR_OPTS_REQ.format(opt="--slave"))

    # option --masters is required (mandatory)
    if not opt.masters:
        parser.error(PARSE_ERR_OPTS_REQ.format(opt="--masters"))

    config_reader = MyDefaultsReader(opt, False)

    # Parse slave connection values
    try:
        slave_vals = parse_connection(opt.slave, config_reader, opt)
    except FormatError:
        _, err, _ = sys.exc_info()
        parser.error("Slave connection values invalid: {0}.".format(err))
    except UtilError:
        _, err, _ = sys.exc_info()
        parser.error("Slave connection values invalid: {0}."
                     "".format(err.errmsg))

    # Parse masters connection values
    masters_vals = []
    masters = opt.masters.split(",")
    if len(masters) == 1:
        parser.error("At least two masters are required for multi-source "
                     "replication.")

    for master in masters:
        try:
            masters_vals.append(parse_connection(master, config_reader, opt))
        except FormatError as err:
            msg = ("Masters connection values invalid or cannot be parsed: "
                   "{0} ({1})".format(master, err))
            raise UtilRplError(msg)
        except UtilError as err:
            msg = ("Masters connection values invalid or cannot be parsed: "
                   "{0} ({1})".format(master, err.errmsg))
            raise UtilRplError(msg)

    # Check hostname alias
    for master_vals in masters_vals:
        if check_hostname_alias(slave_vals, master_vals):
            parser.error("The master and slave are the same host and port.")

    # Check the daemon options
    if opt.daemon:
        # Check if a POSIX system
        if os.name != "posix":
            parser.error("Running mysqlfailover with --daemon is only "
                         "available for POSIX systems.")

        # Check the presence of --log
        if opt.daemon != "stop" and not opt.log_file:
            parser.error("The option --log is required when using --daemon.")

        # Test pidfile
        if opt.daemon != "nodetach":
            pidfile = opt.pidfile or "./rplms_daemon.pid"
            pidfile = os.path.realpath(os.path.normpath(pidfile))
            if opt.daemon == "start":
                # Test if pidfile exists
                if os.path.exists(pidfile):
                    parser.error("pidfile {0} already exists. The daemon is "
                                 "already running?".format(pidfile))
                # Test if pidfile is writable
                try:
                    with open(pidfile, "w") as f:
                        f.write("{0}\n".format(0))
                    # Delete temporary pidfile
                    os.remove(pidfile)
                except IOError as err:
                    parser.error("Unable to write pidfile: {0}"
                                 "".format(err.strerror))
            else:
                # opt.daemon == stop/restart, test if pidfile is readable
                pid = None
                try:
                    if not os.path.exists(pidfile):
                        parser.error("pidfile {0} does not exist."
                                     "".format(pidfile))
                    with open(pidfile, "r") as f:
                        pid = int(f.read().strip())
                except IOError:
                    pid = None
                except ValueError:
                    pid = None
                # Test pid presence
                if not pid:
                    parser.error("Can not read pid from pidfile.")

    if opt.pidfile and not opt.daemon:
        parser.error("The option --daemon is required when using --pidfile.")

    # Purge log file of old data
    if opt.log_file is not None and not purge_log(opt.log_file, opt.log_age):
        parser.error("Error purging log file.")

    # Setup log file
    try:
        logging.basicConfig(filename=opt.log_file, level=logging.INFO,
                            format="%(asctime)s %(levelname)s %(message)s",
                            datefmt=DATE_FORMAT)
    except IOError:
        _, e, _ = sys.exc_info()
        parser.error("Error opening log file: {0}".format(str(e.args[1])))

    # Create dictionary of options
    options = {
        "verbosity": opt.verbosity,
        "quiet": opt.quiet,
        "interval": opt.interval,
        "switchover_interval": opt.switchover_interval,
        "from_beginning": opt.from_beginning,
        "report_values": opt.report_values,
        'format': opt.format,
        "rpl_user": opt.rpl_user,
        "daemon": opt.daemon,
        "pidfile": opt.pidfile,
        "logging": opt.log_file is not None,
        "log_file": opt.log_file,
    }

    try:
        start_ms_replication(slave_vals, masters_vals, options)
    except UtilError:
        _, e, _ = sys.exc_info()
        errmsg = e.errmsg.strip(" ")
        if opt.log_file:
            logging.log(logging.CRITICAL, errmsg)
        print("ERROR: {0}".format(errmsg))
        sys.exit(1)

    sys.exit(0)



#
# Copyright (c) 2010, 2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
This module contains abstractions of a MySQL table and an index.
"""

import multiprocessing
import sys

from mysql.utilities.exception import UtilError, UtilDBError
from mysql.connector.conversion import MySQLConverter
from mysql.utilities.common.format import print_list
from mysql.utilities.common.database import Database
from mysql.utilities.common.lock import Lock
from mysql.utilities.common.server import Server
from mysql.utilities.common.sql_transform import (convert_special_characters,
                                                  quote_with_backticks,
                                                  remove_backtick_quoting,
                                                  is_quoted_with_backticks)

# Constants
_MAXPACKET_SIZE = 1024 * 1024
_MAXBULK_VALUES = 25000
_MAXTHREADS_INSERT = 6
_MAXROWS_PER_THREAD = 100000
_MAXAVERAGE_CALC = 100

_FOREIGN_KEY_QUERY = """
  SELECT CONSTRAINT_NAME, COLUMN_NAME, REFERENCED_TABLE_SCHEMA,
         REFERENCED_TABLE_NAME, REFERENCED_COLUMN_NAME
  FROM INFORMATION_SCHEMA.KEY_COLUMN_USAGE
  WHERE TABLE_SCHEMA = '%s' AND TABLE_NAME = '%s' AND
        REFERENCED_TABLE_SCHEMA IS NOT NULL
"""


class Index(object):
    """
    The Index class encapsulates an index for a given table as defined by
    the output of SHOW INDEXES FROM. The class has the following
    capabilities:

        - Check for duplicates
        - Create DROP statement for index
        - Print index CREATE statement
    """

    def __init__(self, db, index_tuple, verbose=False):
        """Constructor

        db[in]             Name of database
        index_tuple[in]    A tuple from the get_tbl_indexes() result set
        verbose[in]        print extra data during operations (optional)
                           default value = False
        """

        # Initialize and save values
        self.db = db
        self.q_db = quote_with_backticks(db)
        self.verbose = verbose
        self.columns = []
        self.table = index_tuple[0]
        self.q_table = quote_with_backticks(index_tuple[0])
        self.unique = not int(index_tuple[1])
        self.name = index_tuple[2]
        self.q_name = quote_with_backticks(index_tuple[2])
        col = (index_tuple[4], index_tuple[7])
        self.columns.append(col)
        self.accept_nulls = True if index_tuple[9] else False
        self.type = index_tuple[10]
        self.compared = False                    # mark as compared for speed
        self.duplicate_of = None                 # saves duplicate index
        if index_tuple[7] > 0:
            self.column_subparts = True          # check subparts e.g. a(20)
        else:
            self.column_subparts = False

    @staticmethod
    def __cmp_columns(col_a, col_b):
        """Compare two columns on name and subpart lengths if present

        col_a[in]          First column to compare
        col_b[in]          Second column to compare

        Returns True if col_a has the same name as col_b and if the
        subparts are col_a.sub <= col_b.sub.
        """

        sz_this = col_a[1]
        sz_that = col_b[1]
        # if column has the same name
        if col_a[0] == col_b[0]:
            # if they both have sub_parts, compare them
            if sz_this and sz_that:
                if sz_this <= sz_that:
                    return True
                else:
                    return False
            # if this index has a sub_part and the other does
            # not, it is potentially redundant
            elif sz_this and sz_that is None:
                return True
            # if neither have sub_parts, it is a match
            elif sz_this is None and sz_that is None:
                return True
        else:
            return False  # no longer a duplicate

    def __check_column_list(self, index):
        """Compare the column list of this index with another

        index[in]          Instance of Index to compare

        Returns True if column list is a subset of index.
        """

#        # Uniqueness counts - can't be duplicate if uniquess differs
#        #                     except for primary keys which are always unique
#        if index.name != "PRIMARY": # or self.name != "PRIMARY":
#            if self.unique != index.unique:
#                return False
        num_cols_this = len(self.columns)
        num_cols_that = len(index.columns)
        num_cols_same = 0
        if self.type == "BTREE":
            i = 0
            while (i < num_cols_this) and (i < num_cols_that):
                if num_cols_same <= i:  # Ensures first N cols are the same
                    if self.__cmp_columns(self.columns[i], index.columns[i]):
                        num_cols_same = num_cols_same + 1
                    else:
                        break
                i = i + 1
        else:  # HASH, RTREE, FULLTEXT
            if (self.type == "FULLTEXT") and (num_cols_this != num_cols_that):
                return False
            i = 0
            while (i < num_cols_this) and (i < num_cols_that):
                if self.__cmp_columns(self.columns[i], index.columns[i]):
                    num_cols_same = num_cols_same + 1
                else:  # Ensures column lists must match
                    num_cols_same = 0
                    break
                i = i + 1
        if (num_cols_same > 0) and (num_cols_this <= num_cols_that):
            return True
        return False

    def is_duplicate(self, index):
        """Compare this index with another

        index[in]          Instance of Index to compare

        Returns True if this index is a subset of the Index presented.
        """

        # Don't compare the same index - no two indexes can have the same name
        if self.name == index.name:
            return False
        else:
            return self.__check_column_list(index)

    def contains_columns(self, col_names):
        """Check if the current index contains the columns of the given index.

        Returns True if it contains all the columns of the given index,
        otherwise False.
        """
        if len(self.columns) < len(col_names):
            # If has less columns than given index it does not contain all.
            return False
        else:
            this_col_names = [col[0] for col in self.columns]
            # Check if all index column are included in current one..
            for col_name in col_names:
                if col_name not in this_col_names:
                    return False  # found one column not included.

        # Pass previous verification; contains all the columns of given index.
        return True

    def add_column(self, column, sub_part, accept_null):
        """Add a column to the list of columns for this index

        column[in]         Column to add
        sub_part[in]       Sub part of colunm e.g. a(20)
        accept_null[in]       True to indicate the column accepts nulls
        """

        col = (column, sub_part)
        if sub_part > 0:
            self.column_subparts = True
        if accept_null:
            self.accept_nulls = True
        self.columns.append(col)

    def get_drop_statement(self):
        """Get the drop statement for this index

        Note: Ignores PRIMARY key indexes.

        Returns the DROP statement for this index.
        """
        if self.name == "PRIMARY":
            return None
        query_str = "ALTER TABLE {db}.{table} DROP INDEX {name}".format(
            db=self.q_db, table=self.q_table, name=self.q_name
        )
        return query_str

    def get_remove_columns_statement(self, col_names):
        """Get the ALTER TABLE statement to remove columns for this index.

        col_names[in]   list of columns names to remove from the index.

        Returns the ALTER TABLE statement (DROP/ADD) to remove the given
        columns names from the index.
        """
        # Create the new columns list for the index.
        idx_cols = [col[0] for col in self.columns if col[0] not in col_names]
        if not idx_cols:
            # Return a DROP statement if no columns are left.
            query_str = "ALTER TABLE {db}.{table} DROP INDEX {name}".format(
                db=self.q_db, table=self.q_table, name=self.q_name
            )
        else:
            # Otherwise, return a DROP/ADD statement with remaining columns.
            idx_cols_str = ', '.join(idx_cols)
            query_str = ("ALTER TABLE {db}.{table} DROP INDEX {name}, "
                         "ADD INDEX {name} ({cols})".format(db=self.q_db,
                                                            table=self.q_table,
                                                            name=self.q_name,
                                                            cols=idx_cols_str))
        return query_str

    def __get_column_list(self, backtick_quoting=True):
        """Get the column list for an index

        This method is used to print the CREATE and DROP statements.

        backtick_quoting[in]    Indicates if the columns names are to be quoted
                                with backticks or not. By default: True.

        Returns a string representing the list of columns for a
        column list. e.g. 'a, b(10), c'
        """
        col_list = []
        for col in self.columns:
            name, sub_part = (col[0], col[1])
            if backtick_quoting:
                name = quote_with_backticks(name)
            if sub_part > 0:
                col_str = "{0}({1})".format(name, sub_part)
            else:
                col_str = name
            col_list.append(col_str)
        return ', '.join(col_list)

    def print_index_sql(self):
        """Print the CREATE INDEX for indexes and ALTER TABLE for a primary key
        """
        if self.name == "PRIMARY":
            print("ALTER TABLE {db}.{table} ADD PRIMARY KEY ({cols})"
                  "".format(db=self.q_db, table=self.q_table,
                            cols=self.__get_column_list()))
        else:
            create_str = ("CREATE {unique}{fulltext}INDEX {name} ON "
                          "{db}.{table} ({cols}) {using}")
            unique_str = 'UNIQUE ' if self.unique else ''
            fulltext_str = 'FULLTEXT ' if self.type == 'FULLTEXT' else ''
            if (self.type == "BTREE") or (self.type == "RTREE"):
                using_str = 'USING {0}'.format(self.type)
            else:
                using_str = ''
            print(create_str.format(unique=unique_str, fulltext=fulltext_str,
                                    name=self.q_name, db=self.q_db,
                                    table=self.q_table,
                                    cols=self.__get_column_list(),
                                    using=using_str))

    def get_row(self, verbosity=0):
        """Return index information as a list of columns for tabular output.
        """
        cols = self.__get_column_list(backtick_quoting=False)
        if verbosity > 0:
            return (self.db, self.table, self.name, self.type, self.unique,
                    self.accept_nulls, cols)
        return (self.db, self.table, self.name, self.type, cols)


class Table(object):
    """
    The Table class encapsulates a table for a given database. The class
    has the following capabilities:

        - Check to see if the table exists
        - Check indexes for duplicates and redundancies
        - Print list of indexes for the table
        - Extract table data
        - Import table data
        - Copy table data
    """

    def __init__(self, server1, name, options=None):
        """Constructor

        server[in]         A Server object
        name[in]           Name of table in the form (db.table)
        options[in]        options for class: verbose, quiet, get_cols,
            quiet     If True, do not print information messages
            verbose   print extra data during operations (optional)
                      (default is False)
            get_cols  If True, get the column metadata on construction
                      (default is False)
        """
        if options is None:
            options = {}
        self.verbose = options.get('verbose', False)
        self.quiet = options.get('quiet', False)
        self.server = server1

        # Keep table identifier considering backtick quotes
        if is_quoted_with_backticks(name):
            self.q_table = name
            self.q_db_name, self.q_tbl_name = Database.parse_object_name(name)
            self.db_name = remove_backtick_quoting(self.q_db_name)
            self.tbl_name = remove_backtick_quoting(self.q_tbl_name)
            self.table = ".".join([self.db_name, self.tbl_name])
        else:
            self.table = name
            self.db_name, self.tbl_name = Database.parse_object_name(name)
            self.q_db_name = quote_with_backticks(self.db_name)
            self.q_tbl_name = quote_with_backticks(self.tbl_name)
            self.q_table = ".".join([self.q_db_name, self.q_tbl_name])
        self.obj_type = "TABLE"
        self.pri_idx = None

        # We store each type of index in a separate list to make it easier
        # to manipulate
        self.btree_indexes = []
        self.hash_indexes = []
        self.rtree_indexes = []
        self.fulltext_indexes = []
        self.unique_not_null_indexes = None
        self.text_columns = []
        self.blob_columns = []
        self.column_format = None
        self.column_names = []
        self.column_name_type = []
        self.q_column_names = []
        self.indexes_q_names = []
        if options.get('get_cols', False):
            self.get_column_metadata()
        self.dest_vals = None
        self.storage_engine = None

        # Get max allowed packet
        res = self.server.exec_query("SELECT @@session.max_allowed_packet")
        if res:
            self.max_packet_size = res[0][0]
        else:
            self.max_packet_size = _MAXPACKET_SIZE
        # Watch for invalid values
        if self.max_packet_size > _MAXPACKET_SIZE:
            self.max_packet_size = _MAXPACKET_SIZE

        self._insert = "INSERT INTO %s.%s VALUES "
        self.query_options = {  # Used for skipping fetch of rows
            'fetch': False
        }

    def exists(self, tbl_name=None):
        """Check to see if the table exists

        tbl_name[in]       table name (db.table)
                           (optional) If omitted, operation is performed
                           on the class instance table name.

        return True = table exists, False = table does not exist
        """

        db, table = (None, None)
        if tbl_name:
            db, table = Database.parse_object_name(tbl_name)
        else:
            db = self.db_name
            table = self.tbl_name
        res = self.server.exec_query("SELECT TABLE_NAME " +
                                     "FROM INFORMATION_SCHEMA.TABLES " +
                                     "WHERE TABLE_SCHEMA = '%s'" % db +
                                     " and TABLE_NAME = '%s'" % table)

        return (res is not None and len(res) >= 1)

    def get_column_metadata(self, columns=None):
        """Get information about the table for the bulk insert operation.

        This method builds lists that describe the metadata of the table. This
        includes lists for:

          column names
          column format for building VALUES clause
          blob fields - for use in generating INSERT/UPDATE for blobs
          text fields - for use in checking for single quotes

        columns[in]        if None, use EXPLAIN else use column list.
        """

        if columns is None:
            columns = self.server.exec_query("explain %s" % self.q_table)
        stop = len(columns)
        self.column_names = []
        self.q_column_names = []
        col_format_values = [''] * stop
        if columns is not None:
            for col in range(0, stop):
                if is_quoted_with_backticks(columns[col][0]):
                    self.column_names.append(
                        remove_backtick_quoting(columns[col][0]))
                    self.q_column_names.append(columns[col][0])
                else:
                    self.column_names.append(columns[col][0])
                    self.q_column_names.append(
                        quote_with_backticks(columns[col][0]))
                col_type = columns[col][1].lower()
                if ('char' in col_type or 'enum' in col_type
                        or 'set' in col_type or 'binary' in col_type):
                    self.text_columns.append(col)
                    col_format_values[col] = "'%s'"
                elif 'blob' in col_type or 'text'in col_type:
                    self.blob_columns.append(col)
                    col_format_values[col] = "%s"
                elif "date" in col_type or "time" in col_type:
                    col_format_values[col] = "'%s'"
                else:
                    col_format_values[col] = "%s"
        self.column_format = "%s%s%s" % \
                             (" (", ', '.join(col_format_values), ")")

    def get_col_names(self, quote_backticks=False):
        """Get column names for the export operation.

        quote_backticks[in]    If True the column names will be quoted with
                               backticks. Default is False.

        Return (list) column names
        """

        if self.column_format is None:
            self.column_names = []
            self.q_column_names = []
            rows = self.server.exec_query("explain {0}".format(self.q_table))
            for row in rows:
                self.column_names.append(row[0])
                self.q_column_names.append(quote_with_backticks(row[0]))

        return self.q_column_names if quote_backticks else self.column_names

    def get_col_names_types(self, quote_backticks=False):
        """Get a list of tuples of column name and type.

        quote_backticks[in]    If True the column name will be quoted with
                               backticks. Default is False.

        Return (list) of touple (column name, type)
        """

        self.column_name_type = []
        rows = self.server.exec_query("explain {0}".format(self.q_table))
        for row in rows:
            if quote_backticks:
                self.column_name_type.append(
                    [quote_with_backticks(row[0])] + list(row[1:])
                )
            else:
                self.column_name_type.append(row)

        return self.column_name_type

    def has_index(self, index_q_name):
        """A method to determine if this table has a determinate index using
        his name.

        index_q_name[in]    the name of the index (must be quoted).

        returns True if this Table has an index with the given name, otherwise
        false.
        """
        if [idx_q_name for idx_q_name in self.indexes_q_names
           if idx_q_name == index_q_name]:
            return True
        return False

    def get_not_null_unique_indexes(self, refresh=False):
        """get all the unique indexes which columns does not accepts null
        values.
        refresh[in] Boolean value used to force the method to read index
                    information directly from the server, instead of using
                    cached values.

        Returns list of indexes.
        """
        # First check if the instance variable exists.
        if self.unique_not_null_indexes is None or refresh:
            # Get the indexes for the table.
            try:
                self.get_indexes()
            except UtilDBError:
                # Table may not exist yet. Happens on import operations.
                pass
            # Now for each of them, check if they are UNIQUE and NOT NULL.
            no_null_idxes = []
            no_null_idxes.extend(
                [idx for idx in self.btree_indexes if not idx.accept_nulls and
                 idx.unique]
            )
            no_null_idxes.extend(
                [idx for idx in self.hash_indexes if not idx.accept_nulls and
                 idx.unique]
            )
            no_null_idxes.extend(
                [idx for idx in self.rtree_indexes if not idx.accept_nulls and
                 idx.unique]
            )
            no_null_idxes.extend(
                [idx for idx in self.fulltext_indexes if not idx.accept_nulls
                 and idx.unique]
            )
            self.unique_not_null_indexes = no_null_idxes

        return self.unique_not_null_indexes

    def _build_update_blob(self, row, new_db, name):
        """Build an UPDATE statement to update blob fields.

        row[in]            a row to process
        new_db[in]         new database name
        name[in]           name of the table

        Returns UPDATE string
        """
        if self.column_format is None:
            self.get_column_metadata()

        blob_insert = "UPDATE %s.%s SET " % (new_db, name)
        where_values = []
        do_commas = False
        has_data = False
        stop = len(row)
        for col in range(0, stop):
            col_name = self.q_column_names[col]
            if col in self.blob_columns:
                if row[col] is not None and len(row[col]) > 0:
                    if do_commas:
                        blob_insert += ", "
                    blob_insert += "%s = " % col_name + "%s" % \
                                   MySQLConverter().quote(row[col])
                    has_data = True
                    do_commas = True
            else:
                # Convert None values to NULL (not '' to NULL)
                if row[col] is None:
                    value = 'NULL'
                else:
                    value = "'{0}'".format(row[col])
                where_values.append("{0} = {1}".format(col_name, value))
        if has_data:
            return "{0} WHERE {1};".format(blob_insert,
                   " AND ".join(where_values))
        return None

    def _build_insert_blob(self, row, new_db, tbl_name):
        """Build an INSERT statement for the given row.

        row[in]                a row to process
        new_db[in]             new database name
        tbl_name[in]           name of the table

        Returns INSERT string.
        """
        if self.column_format is None:
            self.get_column_metadata()

        converter = MySQLConverter()
        row_vals = []
        # Deal with blob, special characters and NULL values.
        for index, column in enumerate(row):
            if index in self.blob_columns:
                row_vals.append(converter.quote(column))
            elif index in self.text_columns:
                if column is None:
                    row_vals.append("NULL")
                else:
                    row_vals.append(convert_special_characters(column))
            else:
                if column is None:
                    row_vals.append("NULL")
                else:
                    row_vals.append(column)

        # Create the insert statement.
        insert_stm = ("INSERT INTO {0}.{1} VALUES {2};"
                      "".format(new_db, tbl_name,
                                self.column_format % tuple(row_vals)))

        # Replace 'NULL' occurrences with NULL values.
        insert_stm = insert_stm.replace("'NULL'", "NULL")

        return insert_stm

    def get_column_string(self, row, new_db, skip_blobs=False):
        """Return a formatted list of column data.

        row[in]            a row to process
        new_db[in]         new database name
        skip_blobs[in]     boolean value, if True, blob columns are skipped

        Returns (string) column list
        """

        if self.column_format is None:
            self.get_column_metadata()

        blob_inserts = []
        values = list(row)
        is_blob_insert = False
        # find if we have some unique column indexes
        unique_indexes = len(self.get_not_null_unique_indexes())
        # If all columns are blobs or there aren't any UNIQUE NOT NULL indexes
        # then rows won't be correctly copied using the update statement,
        # so we must use insert statements instead.
        if not skip_blobs and (len(self.blob_columns) == len(self.column_names)
                               or self.blob_columns and not unique_indexes):
            blob_inserts.append(self._build_insert_blob(row, new_db,
                                                        self.q_tbl_name))
            is_blob_insert = True
        else:
            # Find blobs
            if self.blob_columns:
                # Save blob updates for later...
                blob = self._build_update_blob(row, new_db, self.q_tbl_name)
                if blob is not None:
                    blob_inserts.append(blob)
                for col in self.blob_columns:
                    values[col] = "NULL"

        if not is_blob_insert:
            # Replace single quotes located in the value for a text field with
            # the correct special character escape sequence. This fixes SQL
            # errors related to using single quotes in a string value that is
            # single quoted. For example, 'this' is it' is changed to
            # 'this\' is it'.
            for col in self.text_columns:
                #Check if the value is not None before replacing quotes
                if values[col]:
                    # Apply escape sequences to special characters
                    values[col] = convert_special_characters(values[col])

            # Build string (add quotes to "string" like types)
            val_str = self.column_format % tuple(values)

            # Change 'None' occurrences with "NULL"
            val_str = val_str.replace(", None", ", NULL")
            val_str = val_str.replace("(None", "(NULL")
            val_str = val_str.replace(", 'None'", ", NULL")
            val_str = val_str.replace("('None'", "(NULL")

        else:
            val_str = None

        return val_str, blob_inserts

    def make_bulk_insert(self, rows, new_db, columns_names=None,
                         skip_blobs=False):
        """Create bulk insert statements for the data

        Reads data from a table (rows) and builds group INSERT statements for
        bulk inserts.

        Note: This method does not print any information to stdout.

        rows[in]           a list of rows to process
        new_db[in]         new database name
        skip_blobs[in]     boolean value, if True, blob columns are skipped

        Returns (tuple) - (bulk insert statements, blob data inserts)
        """

        if self.column_format is None:
            self.get_column_metadata()

        data_inserts = []
        blob_inserts = []
        row_count = 0
        data_size = 0
        val_str = None

        for row in rows:
            if row_count == 0:
                if columns_names:
                    insert_str = "INSERT INTO {0}.{1} ({2}) VALUES ".format(
                        new_db, self.q_tbl_name, ", ".join(columns_names)
                    )
                else:
                    insert_str = self._insert % (new_db, self.q_tbl_name)
                if val_str:
                    row_count += 1
                    insert_str += val_str
                data_size = len(insert_str)

            col_data = self.get_column_string(row, new_db, skip_blobs)
            if len(col_data[1]) > 0:
                blob_inserts.extend(col_data[1])
            if col_data[0]:
                val_str = col_data[0]

                row_size = len(val_str)
                next_size = data_size + row_size + 3
                if ((row_count >= _MAXBULK_VALUES) or
                        (next_size > (int(self.max_packet_size) - 512))):
                   # add to buffer
                    data_inserts.append(insert_str)
                    row_count = 0
                else:
                    row_count += 1
                    if row_count > 1:
                        insert_str += ", "
                    insert_str += val_str
                    data_size += row_size + 3

        if row_count > 0:
            data_inserts.append(insert_str)

        return data_inserts, blob_inserts

    def get_storage_engine(self):
        """Get the storage engine (in UPPERCASE) for the table.

        Returns the name in UPPERCASE of the storage engine use for the table
        or None if the information is not found.
        """
        self.server.exec_query("USE {0}".format(self.q_db_name),
                               self.query_options)
        res = self.server.exec_query(
            "SHOW TABLE STATUS WHERE name = '{0}'".format(self.tbl_name)
        )
        try:
            # Return store engine converted to UPPER cases.
            return res[0][1].upper() if res[0][1] else None
        except IndexError:
            # Return None if table status information is not available.
            return None

    def get_segment_size(self, num_conn=1):
        """Get the segment size based on number of connections (threads).

        num_conn[in]       Number of threads(connections) to use
                           Default = 1 (one large segment)

        Returns (int) segment_size

                Note: if num_conn <= 1 - returns number of rows
        """

        # Get number of rows
        num_rows = 0
        try:
            res = self.server.exec_query("USE %s" % self.q_db_name,
                                         self.query_options)
        except:
            pass
        res = self.server.exec_query("SHOW TABLE STATUS LIKE '%s'" %
                                     self.tbl_name)
        if res:
            num_rows = int(res[0][4])

        if num_conn <= 1:
            return num_rows

        # Calculate number of threads and segment size to fetch
        thread_limit = num_conn
        if thread_limit > _MAXTHREADS_INSERT:
            thread_limit = _MAXTHREADS_INSERT
        if num_rows > (_MAXROWS_PER_THREAD * thread_limit):
            max_threads = thread_limit
        else:
            max_threads = int(num_rows / _MAXROWS_PER_THREAD)
        if max_threads == 0:
            max_threads = 1
        if max_threads > 1 and self.verbose:
            print "# Using multi-threaded insert option. Number of " \
                  "threads = %d." % max_threads
        return (num_rows / max_threads) + max_threads

    def _bulk_insert(self, rows, new_db, destination=None):
        """Import data using bulk insert

        Reads data from a table and builds group INSERT statements for writing
        to the destination server specified (new_db.name).

        This method is designed to be used in a thread for parallel inserts.
        As such, it requires its own connection to the destination server.

        Note: This method does not print any information to stdout.

        rows[in]           a list of rows to process
        new_db[in]         new database name
        destination[in]    the destination server
        """
        if self.dest_vals is None:
            self.dest_vals = self.get_dest_values(destination)

        # Spawn a new connection
        server_options = {
            'conn_info': self.dest_vals,
            'role': "thread",
        }
        dest = Server(server_options)
        dest.connect()

        # Issue the write lock
        lock_list = [("%s.%s" % (new_db, self.q_tbl_name), 'WRITE')]
        my_lock = Lock(dest, lock_list, {'locking': 'lock-all', })

        # First, turn off foreign keys if turned on
        dest.disable_foreign_key_checks(True)

        if self.column_format is None:
            self.get_column_metadata()

        data_lists = self.make_bulk_insert(rows, new_db)
        insert_data = data_lists[0]
        blob_data = data_lists[1]

        # Insert the data first
        for data_insert in insert_data:
            try:
                dest.exec_query(data_insert, self.query_options)
            except UtilError, e:
                raise UtilError("Problem inserting data. "
                                "Error = %s" % e.errmsg)

        # Now insert the blob data if there is any
        for blob_insert in blob_data:
            try:
                dest.exec_query(blob_insert, self.query_options)
            except UtilError, e:
                raise UtilError("Problem updating blob field. "
                                "Error = %s" % e.errmsg)

        # Now, turn on foreign keys if they were on at the start
        dest.disable_foreign_key_checks(False)
        my_lock.unlock()
        del dest

    def insert_rows(self, rows, new_db, destination=None, spawn=False):
        """Insert rows in the table using bulk copy.

        This method opens a new connect to the destination server to insert
        the data with a bulk copy. If spawn is True, the method spawns a new
        process and returns it. This allows for using a multi-threaded insert
        which can be faster on some platforms. If spawn is False, the method
        will open a new connection to insert the data.

        num_conn[in]       Number of threads(connections) to use for insert
        rows[in]           List of rows to insert
        new_db[in]         Rename the db to this name
        destination[in]    Destination server
                           Default = None (copy to same server)
        spawn[in]          If True, spawn a new process for the insert
                           Default = False

        Returns If spawn == True, process
                If spawn == False, None
        """

        if self.column_format is None:
            self.get_column_metadata()

        if self.dest_vals is None:
            self.dest_vals = self.get_dest_values(destination)

        proc = None
        if spawn:
            proc = multiprocessing.Process(target=self._bulk_insert,
                                           args=(rows, new_db, destination))
        else:
            self._bulk_insert(rows, new_db, destination)

        return proc

    def _clone_data(self, new_db):
        """Clone table data.

        This method will copy all of the data for a table
        from the old database to the new database on the same server.

        new_db[in]         New database name for the table
        """
        query_str = "INSERT INTO %s.%s SELECT * FROM %s.%s" % \
                    (new_db, self.q_tbl_name, self.q_db_name, self.q_tbl_name)
        if self.verbose and not self.quiet:
            print query_str

        # Disable foreign key checks to allow data to be copied without running
        # into foreign key referential integrity issues
        self.server.disable_foreign_key_checks(True)
        self.server.exec_query(query_str)
        self.server.disable_foreign_key_checks(False)

    def copy_data(self, destination, cloning=False, new_db=None,
                  connections=1):
        """Retrieve data from a table and copy to another server and database.

        Reads data from a table and inserts the correct INSERT statements into
        the file provided.

        Note: if connections < 1 - retrieve the data one row at-a-time

        destination[in]    Destination server
        cloning[in]        If True, we are copying on the same server
        new_db[in]         Rename the db to this name
        connections[in]    Number of threads(connections) to use for insert
        """

        if new_db is None:
            new_db = self.q_db_name
        else:
            # If need quote new_db identifier with backticks
            if not is_quoted_with_backticks(new_db):
                new_db = quote_with_backticks(new_db)

        num_conn = int(connections)

        if cloning:
            self._clone_data(new_db)
        else:
            # Read and copy the data
            pthreads = []
            for rows in self.retrieve_rows(num_conn):
                p = self.insert_rows(rows, new_db, destination, num_conn > 1)
                if p is not None:
                    p.start()
                    pthreads.append(p)

            if num_conn > 1:
                # Wait for all threads to finish
                for p in pthreads:
                    p.join()

    def retrieve_rows(self, num_conn=1):
        """Retrieve the table data in rows.

        This method can be used to retrieve rows from a table as a generator
        specifying how many rows to retrieve at one time (segment_size is
        calculated based on number of rows / number of connections).

        Note: if num_conn < 1 - retrieve the data one row at-a-time

        num_conn[in]       Number of threads(connections) to use
                           Default = 1 (one large segment)

        Returns (yield) row data
        """

        if num_conn > 1:
            # Only get the segment size when needed.
            segment_size = self.get_segment_size(num_conn)

        # Execute query to get all of the data
        cur = self.server.exec_query("SELECT * FROM {0}".format(self.q_table),
                                     self.query_options)

        while True:
            rows = None
            if num_conn < 1:
                rows = []
                row = cur.fetchone()
                if row is None:
                    raise StopIteration()
                rows.append(row)
                #print "ROWS 1:", rows
            elif num_conn == 1:
                rows = cur.fetchall()
                #print "ROWS 2:", rows
                yield rows
                raise StopIteration()
            else:
                rows = cur.fetchmany(segment_size)
                if not rows:
                    raise StopIteration()
                #print "ROWS 3:", rows
            if rows is None:
                raise StopIteration()
            yield rows

        cur.close()

    def get_dest_values(self, destination=None):
        """Get the destination connection values if not already set.

        destination[in]    Connection values for destination server

        Returns connection values for destination if set or self.server
        """
        # Get connection to database
        if destination is None:
            conn_val = {
                "host": self.server.host,
                "user": self.server.user,
                "passwd": self.server.passwd,
                "unix_socket": self.server.socket,
                "port": self.server.port
            }
        else:
            conn_val = {
                "host": destination.host,
                "user": destination.user,
                "passwd": destination.passwd,
                "unix_socket": destination.socket,
                "port": destination.port
            }
        return conn_val

    def get_tbl_indexes(self):
        """Return a result set containing all indexes for a given table

        Returns result set
        """
        res = self.server.exec_query("SHOW INDEXES FROM %s" % self.q_table)
        return res

    def get_tbl_foreign_keys(self):
        """Return a result set containing all foreign keys for the table

        Returns result set
        """
        res = self.server.exec_query(_FOREIGN_KEY_QUERY % (self.db_name,
                                                           self.tbl_name))
        return res

    @staticmethod
    def __append(indexes, index):
        """Encapsulated append() method to ensure the primary key index
        is placed at the front of the list.
        """

        # Put the primary key first so that it can be compared to all indexes
        if index.name == "PRIMARY":
            indexes.insert(0, index)
        else:
            indexes.append(index)

    @staticmethod
    def __check_index(index, indexes, master_list):
        """Check a single index for duplicate or redundancy against a list
        of other Indexes.

        index[in]          The Index to compare
        indexes[in]        A list of Index instances to compare
        master_list[in]    A list of know duplicate Index instances

        Returns a tuple of whether duplicates are found and if found the
        list of duplicate indexes for this table
        """

        duplicates_found = False
        duplicate_list = []
        if indexes and index:
            for idx in indexes:
                if index == idx:
                    continue
                # Don't compare b == a when a == b has already occurred
                if not index.compared and idx.is_duplicate(index):
                    # make sure we haven't already found this match
                    if not idx.column_subparts:
                        idx.compared = True
                    if idx not in master_list:
                        duplicates_found = True
                        # PRIMARY key can be identified as redundant of an
                        # unique index with more columns, in that case always
                        # mark the other as the duplicate.
                        if idx.name == "PRIMARY":
                            index.duplicate_of = idx
                            duplicate_list.append(index)
                        else:
                            idx.duplicate_of = index
                            duplicate_list.append(idx)
        return (duplicates_found, duplicate_list)

    def __check_index_list(self, indexes):
        """Check a list of Index instances for duplicates.

        indexes[in]        A list of Index instances to compare

        Returns a tuple of whether duplicates are found and if found the
        list of duplicate indexes for this table
        """

        duplicates_found = False
        duplicate_list = []
        # Caller must ensure there are at least 2 elements in the list.
        if len(indexes) < 2:
            return (False, None)
        for index in indexes:
            res = self.__check_index(index, indexes, duplicate_list)
            if res[0]:
                duplicates_found = True
                duplicate_list.extend(res[1])
        return (duplicates_found, duplicate_list)

    def __check_clustered_index_list(self, indexes):
        """ Check for indexes containing the clustered index from the list.

        indexes[in]     list of indexes instances to check.

        Returns the list of indexes that contain the clustered index or
        None (if none found).
        """
        redundant_indexes = []
        if not self.pri_idx:
            self.get_primary_index()
        pri_idx_cols = [col[0] for col in self.pri_idx]
        for index in indexes:
            if index.name == 'PRIMARY':
                # Skip primary key.
                continue
            elif index.contains_columns(pri_idx_cols):
                redundant_indexes.append(index)

        return redundant_indexes if redundant_indexes else []

    def _get_index_list(self):
        """Get the list of indexes for a table.
        Returns list containing indexes.
        """
        rows = self.get_tbl_indexes()
        return rows

    def get_primary_index(self):
        """Retrieve the primary index columns for this table.
        """
        pri_idx = []

        rows = self.server.exec_query("EXPLAIN {0}".format(self.q_table))

        # Return False if no indexes found.
        if not rows:
            return pri_idx

        for row in rows:
            if row[3] == 'PRI':
                pri_idx.append(row)

        self.pri_idx = pri_idx

        return pri_idx

    def get_column_explanation(self, column_name):
        """Retrieve the explain description for the given column.
        """
        column_exp = []

        rows = self.server.exec_query("EXPLAIN {0}".format(self.q_table))

        # Return False if no indexes found.
        if not rows:
            return column_exp

        for row in rows:
            if row[0] == column_name:
                column_exp.append(row)

        return column_exp

    def get_indexes(self):
        """Retrieve the indexes from the server and load them into lists
        based on type.

        Returns True - table has indexes, False - table has no indexes
        """

        self.btree_indexes = []
        self.hash_indexes = []
        self.rtree_indexes = []
        self.fulltext_indexes = []
        self.indexes_q_names = []

        if self.verbose:
            print "# Getting indexes for %s" % (self.table)
        rows = self._get_index_list()

        # Return False if no indexes found.
        if not rows:
            return False
        idx = None
        prev_name = ""
        for row in rows:
            if (row[2] != prev_name) or (prev_name == ""):
                prev_name = row[2]
                idx = Index(self.db_name, row)
                if idx.type == "BTREE":
                    self.__append(self.btree_indexes, idx)
                elif idx.type == "HASH":
                    self.__append(self.hash_indexes, idx)
                elif idx.type == "RTREE":
                    self.__append(self.rtree_indexes, idx)
                else:
                    self.__append(self.fulltext_indexes, idx)
            elif idx:
                idx.add_column(row[4], row[7], row[9])
            self.indexes_q_names.append(quote_with_backticks(row[2]))
        return True

    def check_indexes(self, show_drops=False):
        """Check for duplicate or redundant indexes and display all matches

        show_drops[in]     (optional) If True the DROP statements are printed

        Note: You must call get_indexes() prior to calling this method. If
        get_indexes() is not called, no duplicates will be found.
        """

        dupes = []
        res = self.__check_index_list(self.btree_indexes)
        # if there are duplicates, add them to the dupes list
        if res[0]:
            dupes.extend(res[1])
        res = self.__check_index_list(self.hash_indexes)
        # if there are duplicates, add them to the dupes list
        if res[0]:
            dupes.extend(res[1])
        res = self.__check_index_list(self.rtree_indexes)
        # if there are duplicates, add them to the dupes list
        if res[0]:
            dupes.extend(res[1])
        # We sort the fulltext index columns - easier to do it once here
        for index in self.fulltext_indexes:
            cols = index.columns
            cols.sort(key=lambda cols: cols[0])
        res = self.__check_index_list(self.fulltext_indexes)
        # if there are duplicates, add them to the dupes list
        if res[0]:
            dupes.extend(res[1])

        # Check if secondary keys contains the clustered index (i.e. Primary
        # key). In InnoDB, each record in a secondary index contains the
        # primary key columns. Therefore the use of keys that include the
        # primary key might be redundant.
        redundant_idxs = []
        if not self.storage_engine:
            self.storage_engine = self.get_storage_engine()
        if self.storage_engine == 'INNODB':
            all_indexes = self.btree_indexes
            all_indexes.extend(self.hash_indexes)
            all_indexes.extend(self.rtree_indexes)
            all_indexes.extend(self.fulltext_indexes)
            redundant_idxs = self.__check_clustered_index_list(all_indexes)

        # Print duplicate and redundant keys on composite indexes.
        if len(dupes) > 0:
            plural_1, verb_conj, plural_2 = (
                ('', 'is a', '') if len(dupes) == 1 else ('es', 'are', 's')
            )
            print("# The following index{0} {1} duplicate{2} or redundant "
                  "for table {3}:".format(plural_1, verb_conj, plural_2,
                                          self.table))
            for index in dupes:
                print("#")
                index.print_index_sql()
                print("#     may be redundant or duplicate of:")
                index.duplicate_of.print_index_sql()
            if show_drops:
                print("#\n# DROP statement{0}:\n#".format(plural_2))
                for index in dupes:
                    print("{0};".format(index.get_drop_statement()))
                print("#")

        # Print redundant indexes containing clustered key.
        if redundant_idxs:
            plural, verb_conj, plural_2 = (
                ('', 's', '') if len(redundant_idxs) == 1 else ('es', '', 's')
            )

            print("# The following index{0} for table {1} contain{2} the "
                  "clustered index and might be redundant:".format(plural,
                                                                   self.table,
                                                                   verb_conj))
            for index in redundant_idxs:
                print("#")
                index.print_index_sql()
            if show_drops:
                print("#\n# DROP/ADD statement{0}:\n#".format(plural_2))
                # Get columns from primary key to be removed.
                pri_idx_cols = [col[0] for col in self.pri_idx]
                for index in redundant_idxs:
                    print("{0};".format(
                        index.get_remove_columns_statement(pri_idx_cols)
                    ))
                print("#")

        if not self.quiet and not dupes and not redundant_idxs:
            print("# Table {0} has no duplicate nor redundant "
                  "indexes.".format(self.table))

    def show_special_indexes(self, fmt, limit, best=False):
        """Display a list of the best or worst queries for this table.

        This shows the best (first n) or worst (last n) performing queries
        for a given table.

        fmt[in]            format out output = sql, table, tab, csv
        limit[in]          number to limit the display
        best[in]           (optional) if True, print best performing indexes
                                      if False, print worst performing indexes
        """

        _QUERY = """
            SELECT
                t.TABLE_SCHEMA AS `db`, t.TABLE_NAME AS `table`,
                s.INDEX_NAME AS `index name`, s.COLUMN_NAME AS `field name`,
                s.SEQ_IN_INDEX `seq in index`, s2.max_columns AS `# cols`,
                s.CARDINALITY AS `card`, t.TABLE_ROWS AS `est rows`,
                ROUND(((s.CARDINALITY / IFNULL(
                IF(t.TABLE_ROWS < s.CARDINALITY, s.CARDINALITY, t.TABLE_ROWS),
                0.01)) * 100), 2) AS `sel_percent`
            FROM INFORMATION_SCHEMA.STATISTICS s
                INNER JOIN INFORMATION_SCHEMA.TABLES t
                ON s.TABLE_SCHEMA = t.TABLE_SCHEMA
                AND s.TABLE_NAME = t.TABLE_NAME
            INNER JOIN (
                SELECT TABLE_SCHEMA, TABLE_NAME, INDEX_NAME,
                    MAX(SEQ_IN_INDEX) AS max_columns
                FROM INFORMATION_SCHEMA.STATISTICS
                WHERE TABLE_SCHEMA = %s AND TABLE_NAME = %s
                      AND INDEX_NAME != 'PRIMARY'
                GROUP BY TABLE_SCHEMA, TABLE_NAME, INDEX_NAME
             ) AS s2
             ON s.TABLE_SCHEMA = s2.TABLE_SCHEMA
                AND s.TABLE_NAME = s2.TABLE_NAME
                AND s.INDEX_NAME = s2.INDEX_NAME
            WHERE t.TABLE_SCHEMA != 'mysql'
                AND t.TABLE_ROWS > 10 /* Only tables with some rows */
                AND s.CARDINALITY IS NOT NULL
                AND (s.CARDINALITY / IFNULL(
                IF(t.TABLE_ROWS < s.CARDINALITY, s.CARDINALITY, t.TABLE_ROWS),
                0.01)) <= 1.00
            ORDER BY `sel_percent`
        """
        query_options = {
            'params': (self.db_name, self.tbl_name,)
        }
        rows = []
        idx_type = "best"
        if not best:
            idx_type = "worst"
        if best:
            rows = self.server.exec_query(_QUERY + "DESC LIMIT %s" % limit,
                                          query_options)
        else:
            rows = self.server.exec_query(_QUERY + "LIMIT %s" % limit,
                                          query_options)
        if rows:
            print("#")
            if limit == 1:
                print("# Showing the {0} performing index from "
                      "{1}:".format(idx_type, self.table))
            else:
                print("# Showing the top {0} {1} performing indexes from "
                      "{2}:".format(limit, idx_type, self.table))
            print("#")
            cols = ("database", "table", "name", "column", "sequence",
                    "num columns", "cardinality", "est. rows", "percent")
            print_list(sys.stdout, fmt, cols, rows)
        else:
            print("# WARNING: Not enough data to calculate "
                  "best/worst indexes.")

    @staticmethod
    def __print_index_list(indexes, fmt, no_header=False, verbosity=0):
        """Print the list of indexes

        indexes[in]        list of indexes to print
        fmt[in]            format out output = sql, table, tab, csv
        no_header[in]      (optional) if True, do not print the header
        """
        if fmt == "sql":
            for index in indexes:
                index.print_index_sql()
        else:
            if verbosity > 0:
                cols = ("database", "table", "name", "type", "unique",
                        "accepts nulls", "columns")
            else:
                cols = ("database", "table", "name", "type", "columns")

            rows = []
            for index in indexes:
                rows.append(index.get_row(verbosity))
            print_list(sys.stdout, fmt, cols, rows, no_header)

    def print_indexes(self, fmt, verbosity):
        """Print all indexes for this table

        fmt[in]         format out output = sql, table, tab, csv
        """

        print "# Showing indexes from %s:\n#" % (self.table)
        if fmt == "sql":
            self.__print_index_list(self.btree_indexes, fmt,
                                    verbosity=verbosity)
            self.__print_index_list(self.hash_indexes, fmt, False,
                                    verbosity=verbosity)
            self.__print_index_list(self.rtree_indexes, fmt, False,
                                    verbosity=verbosity)
            self.__print_index_list(self.fulltext_indexes, fmt, False,
                                    verbosity=verbosity)
        else:
            master_indexes = []
            master_indexes.extend(self.btree_indexes)
            master_indexes.extend(self.hash_indexes)
            master_indexes.extend(self.rtree_indexes)
            master_indexes.extend(self.fulltext_indexes)
            self.__print_index_list(master_indexes, fmt,
                                    verbosity=verbosity)
        print "#"

    def has_primary_key(self):
        """Check to see if there is a primary key.
        Returns bool - True - a primary key was found,
                       False - no primary key.
        """
        primary_key = False
        rows = self._get_index_list()
        for row in rows:
            if row[2] == "PRIMARY":
                primary_key = True
        return primary_key

    def has_unique_key(self):
        """Check to see if there is a unique key.
        Returns bool - True - a unique key was found,
                       False - no unique key.
        """
        unique_key = False
        rows = self._get_index_list()
        for row in rows:
            if row[1] == '0':
                unique_key = True
        return unique_key



#
# Copyright (c) 2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
This file contains features to check the data consistency in a replication
topology (i.e., between the master and its slaves, or only slaves), providing
synchronization features to perform the check over the (supposed) same data of
a system with replication active (running).
"""
import re
import sys

from multiprocessing.pool import ThreadPool

from mysql.utilities.command.dbcompare import diff_objects, get_common_objects
from mysql.utilities.common.database import Database
from mysql.utilities.common.messages import ERROR_USER_WITHOUT_PRIVILEGES
from mysql.utilities.common.pattern_matching import convertSQL_LIKE2REGEXP
from mysql.utilities.common.replication import (get_last_server_gtid,
                                                gtid_set_cardinality,
                                                gtid_set_union)
from mysql.utilities.common.sql_transform import quote_with_backticks
from mysql.utilities.common.topology import Topology
from mysql.utilities.common.user import User
from mysql.utilities.exception import UtilError

# Regular expression to handle the server version format.
_RE_VERSION_FORMAT = r'^(\d+\.\d+(\.\d+)*).*$'


class RPLSynchronizer(object):
    """Class to manage the features of the replication synchronization checker.

    The RPLSynchronizer class is used to manage synchronization check between
    servers of a replication topology, namely between the master and its
    slaves or only between slaves. It provides functions to determine the
    slaves missing transactions (i.e., missing GTIDs) and check data
    consistency.
    """

    def __init__(self, master_cnx_dic, slaves_cnx_dic_lst, options):
        """Constructor.

        options[in]       dictionary of options (e.g., discover, timeouts,
                          verbosity).
        """
        self._verbosity = options.get('verbosity')
        self._rpl_timeout = options.get('rpl_timeout')
        self._checksum_timeout = options.get('checksum_timeout')
        self._interval = options.get('interval')

        self._rpl_topology = Topology(master_cnx_dic, slaves_cnx_dic_lst,
                                      options)
        self._slaves = self._rpl_topology.get_slaves_dict()

        # Set base server used as reference for comparisons.
        self._base_server = None
        self._base_server_key = None
        self._set_base_server()

        # Check user permissions to perform the consistency check.
        self._check_privileges()

        # Check usage of replication filters.
        self._master_rpl_filters = {}
        self._slaves_rpl_filters = {}
        self._check_rpl_filters()

    def _set_base_server(self):
        """Set the base server used for comparison in the internal state.

        Set the master if used or the first slave from the topology as the
        base server. The base server is the one used as a reference for
        comparison with the others. This method sets two instance variables:
        _base_server with the Server instance, and _base_server_key with the
        string identifying the server (format: 'host@port').

        Note: base server might need to be changed (set again) if it is
        removed from the topology for some reason (e.g. GTID disabled).
        """
        master = self._get_master()
        self._base_server = master if master \
            else self._rpl_topology.slaves[0]['instance']
        self._base_server_key = "{0}@{1}".format(self._base_server.host,
                                                 self._base_server.port)

    def _get_slave(self, slave_key):
        """Get the slave server instance for the specified key 'host@port'.

        This function retrieves the Server instance of for a slave from the
        internal state by specifying the key that uniquely identifies it,
        i.e. 'host@port'.

        slave_key[in]   String with the format 'host@port' that uniquely
                        identifies a server.

        Returns a Server instance of the slave with the specified key value
        (i.e., 'host@port').
        """
        slave_dict = self._slaves[slave_key]
        return slave_dict['instance']

    def _get_master(self):
        """Get the master server instance.

        This function retrieves the Server instance of the master (in the
        replication topology).

        Returns a Server instance of the master.
        """
        return self._rpl_topology.master

    def _check_privileges(self):
        """Check required privileges to perform the synchronization check.

        This method check if the used users for the master and slaves possess
        the required privileges to perform the synchronization check. More
        specifically, the following privileges are required:
            - on the master: SUPER or REPLICATION CLIENT, LOCK TABLES and
                             SELECT;
            - on slaves: SUPER and SELECT.
        An exception is thrown if users doesn't have enough privileges.
        """
        if self._verbosity:
            print("# Checking users permission to perform consistency check.\n"
                  "#")

        # Check privileges for master.
        master_priv = [('SUPER', 'REPLICATION CLIENT'), ('LOCK TABLES',),
                       ('SELECT',)]
        master_priv_str = "SUPER or REPLICATION CLIENT, LOCK TABLES and SELECT"
        if self._get_master():
            server = self._get_master()
            user_obj = User(server, "{0}@{1}".format(server.user, server.host))
            for any_priv_tuple in master_priv:
                has_privilege = any(
                    [user_obj.has_privilege('*', '*', priv)
                        for priv in any_priv_tuple]
                )
                if not has_privilege:
                    raise UtilError(ERROR_USER_WITHOUT_PRIVILEGES.format(
                        user=server.user, host=server.host, port=server.port,
                        operation='the synchronization check',
                        req_privileges=master_priv_str
                    ))

        # Check privileges for slaves.
        slave_priv = [('SUPER',), ('SELECT',)]
        slave_priv_str = "SUPER and SELECT"
        for slave_key in self._slaves:
            server = self._get_slave(slave_key)
            user_obj = User(server, "{0}@{1}".format(server.user, server.host))
            for any_priv_tuple in slave_priv:
                has_privilege = any(
                    [user_obj.has_privilege('*', '*', priv)
                        for priv in any_priv_tuple]
                )
                if not has_privilege:
                    raise UtilError(
                        "User '{0}' on '{1}@{2}' does not have sufficient "
                        "privileges to perform the synchronization check "
                        "(required: {3}).".format(server.user, server.host,
                                                  server.port, slave_priv_str)
                    )

    def _check_rpl_filters(self):
        """Check usage of replication filters.

        Check the usage of replication filtering option on the master (if
        defined) and slaves, and set the internal state with the found options
        (to check later).
        """
         # Get binlog filtering option for the master.
        if self._get_master():
            m_filters = self._get_master().get_binlog_exceptions()
            if m_filters:
                # Set filtering option for master.
                self._master_rpl_filters['binlog_do_db'] = \
                    m_filters[0][1].split(',') if m_filters[0][1] else None
                self._master_rpl_filters['binlog_ignore_db'] = \
                    m_filters[0][2].split(',') if m_filters[0][2] else None

        # Get replication filtering options for each slave.
        for slave_key in self._slaves:
            slave = self._get_slave(slave_key)
            s_filters = slave.get_slave_rpl_filters()
            if s_filters:
                # Handle known server issues with some replication filters,
                # leading to inconsistent GTID sets. Sync not supported for
                # server with those issues.
                issues = [(0, 'replicate_do_db'), (1, 'replicate_ignore_db'),
                          (4, 'replicate_wild_do_table')]
                for index, rpl_opt in issues:
                    if s_filters[index]:
                        raise UtilError(
                            "Use of {0} option is not supported. There is a "
                            "known issue with the use this replication filter "
                            "and GTID for some server versions. Issue "
                            "detected for '{1}'.".format(rpl_opt, slave_key))
                # Set map (dictionary) with the slave filtering options.
                filters_map = {
                    'replicate_do_db':
                    s_filters[0].split(',') if s_filters[0] else None,
                    'replicate_ignore_db':
                    s_filters[1].split(',') if s_filters[1] else None,
                    'replicate_do_table':
                    s_filters[2].split(',') if s_filters[2] else None,
                    'replicate_ignore_table':
                    s_filters[3].split(',') if s_filters[3] else None,
                }
                # Handle wild-*-table filters differently to create
                # corresponding regexp.
                if s_filters[4]:
                    wild_list = s_filters[4].split(',')
                    filters_map['replicate_wild_do_table'] = wild_list
                    # Create auxiliary list with compiled regexp to match.
                    regexp_list = []
                    for wild in wild_list:
                        regexp = re.compile(convertSQL_LIKE2REGEXP(wild))
                        regexp_list.append(regexp)
                    filters_map['regexp_do_table'] = regexp_list
                else:
                    filters_map['replicate_wild_do_table'] = None
                    filters_map['regexp_do_table'] = None
                if s_filters[5]:
                    wild_list = s_filters[5].split(',')
                    filters_map['replicate_wild_ignore_table'] = wild_list
                    # Create auxiliary list with compiled regexp to match.
                    regexp_list = []
                    for wild in wild_list:
                        regexp = re.compile(convertSQL_LIKE2REGEXP(wild))
                        regexp_list.append(regexp)
                    filters_map['regexp_ignore_table'] = regexp_list
                else:
                    filters_map['replicate_wild_ignore_table'] = None
                    filters_map['regexp_ignore_table'] = None
                # Set filtering options for the slave.
                self._slaves_rpl_filters[slave_key] = filters_map

        # Print warning if filters are found.
        if self._master_rpl_filters or self._slaves_rpl_filters:
            print("# WARNING: Replication filters found on checked "
                  "servers. This can lead data consistency issues "
                  "depending on how statements are evaluated.\n"
                  "# More information: "
                  "http://dev.mysql.com/doc/en/replication-rules.html")
            if self._verbosity:
                # Print filter options in verbose mode.
                if self._master_rpl_filters:
                    print("# Master '{0}@{1}':".format(
                        self._get_master().host, self._get_master().port
                    ))
                    for rpl_filter in self._master_rpl_filters:
                        if self._master_rpl_filters[rpl_filter]:
                            print("#   - {0}: {1}".format(
                                rpl_filter,
                                ', '.join(
                                    self._master_rpl_filters[rpl_filter]
                                )
                            ))
                if self._slaves_rpl_filters:
                    for slave_key in self._slaves_rpl_filters:
                        print("# Slave '{0}':".format(slave_key))
                        filters_map = self._slaves_rpl_filters[slave_key]
                        for rpl_filter in filters_map:
                            if (rpl_filter.startswith('replicate')
                                    and filters_map[rpl_filter]):
                                print("#   - {0}: {1}".format(
                                    rpl_filter,
                                    ', '.join(filters_map[rpl_filter])
                                ))

    def _is_rpl_filtered(self, db_name, tbl_name=None, slave=None):
        """ Check if the given object is to be filtered by replication.

        This method checks if the given database or table name is
        supposed to be filtered by replication (i.e., not replicated),
        according to the defined replication filters for the master or
        the specified slave.

        db_name[in]     Name of the database to check (not backtick quoted) or
                        associated to the table to check..
        tbl_name[in]    Name of the table to check (not backtick quoted).
                        Table level filtering rules are only checked if this
                        value is not None. By default None, meaning that only
                        the database level rules are checked.
        slave[in]       Identification of the slave in the format 'host@port'
                        to check, determining which filtering rules will be
                        checked. If None only the master filtering rules are
                        checked, otherwise the rule of the specified slaves
                        are used. By default: None.

        Returns a boolean value indicating if the given database or table is
        supposed to be filtered by the replication  or not. More precisely,
        if True then updates associated to the object are (supposedly) not
        replicated, otherwise they are replicated.
        """
        def match_regexp(name, regex_list):
            """ Check if 'name' matches one of the regex in the given list.
            """
            for regex in regex_list:
                if regex.match(name):
                    return True
            return False

        # Determine object to check and set full qualified name.
        is_db = tbl_name is None
        obj_name = db_name if is_db else '{0}.{1}'.format(db_name, tbl_name)

        # Match replication filter for Master.
        if not slave and is_db and self._master_rpl_filters:
            if self._master_rpl_filters['binlog_do_db']:
                if obj_name in self._master_rpl_filters['binlog_do_db']:
                    return False
                else:
                    return True
            elif self._master_rpl_filters['binlog_ignore_db']:
                if obj_name in self._master_rpl_filters['binlog_ignore_db']:
                    return True

        # Match replication filters for the specified slave.
        if slave and slave in self._slaves_rpl_filters:
            rpl_filter = self._slaves_rpl_filters[slave]
            if is_db:
                if rpl_filter['replicate_do_db']:
                    if obj_name in rpl_filter['replicate_do_db']:
                        return False
                    else:
                        return True
                elif (rpl_filter['replicate_ignore_db']
                        and obj_name in rpl_filter['replicate_ignore_db']):
                    return True
            else:
                if (rpl_filter['replicate_do_table']
                        and obj_name in rpl_filter['replicate_do_table']):
                    return False
                if (rpl_filter['replicate_ignore_table']
                        and obj_name in rpl_filter['replicate_ignore_table']):
                    return True
                if (rpl_filter['replicate_wild_do_table']
                        and match_regexp(obj_name,
                                         rpl_filter['regexp_do_table'])):
                    return False
                if (rpl_filter['replicate_wild_ignore_table']
                        and match_regexp(obj_name,
                                         rpl_filter['regexp_ignore_table'])):
                    return True
                if (rpl_filter['replicate_do_table']
                        or rpl_filter['replicate_wild_do_table']):
                    return True

        # Do not filter replication for object (if no filter rule matched).
        return False

    def _apply_for_all_slaves(self, slaves, function, args=(), kwargs=None,
                              multithreading=False):
        """Apply specified function to all given slaves.

        This function allow the execution (concurrently or not) of the
        specified function with the given arguments on all the specified
        slaves.

        slaves[in]          List of slaves to apply the function. It is assumed
                            that the list is composed by strings with the
                            format 'host@port', identifying each slave.
        function[in]        Name of the function (string) to apply on all
                            slaves.
        args[in]            Tuple with all the function arguments (except
                            keyword arguments).
        kwargs[in]          Dictionary with all the function keyword arguments.
        multithreading[in]  Boolean value indicating if the function will be
                            applied concurrently on all slaves. By default
                            False, no concurrency.

        Return a list of tuples composed by two elements: a string identifying
        the slave ('host@port') and the result of the execution of the target
        function for the corresponding slave.
        """
        if kwargs is None:
            kwargs = {}
        if multithreading:
            # Create a pool of threads to execute the method for each slave.
            pool = ThreadPool(processes=len(slaves))
            thread_res_lst = []
            for slave_key in slaves:
                slave = self._get_slave(slave_key)
                thread_res = pool.apply_async(getattr(slave, function), args,
                                              kwargs)
                thread_res_lst.append((slave_key, thread_res))
            pool.close()
            # Wait for all threads to finish here to avoid RuntimeErrors when
            # waiting for the result of a thread that is already dead.
            pool.join()
            # Get the result from each slave and return the results.
            res = []
            for slave_key, thread_res in thread_res_lst:
                res.append((slave_key, thread_res.get()))
            return res
        else:
            res = []
            for slave_key in slaves:
                slave = self._get_slave(slave_key)
                slave_res = getattr(slave, function)(*args, **kwargs)
                res.append((slave_key, slave_res))
            return res

    def check_server_versions(self):
        """Check server versions.

        Check all server versions and report version differences.
        """
        srv_versions = {}
        # Get the server version of the master if used.
        master = self._get_master()
        if master:
            master_version = master.get_version()
            match = re.match(_RE_VERSION_FORMAT, master_version.strip())
            if match:
                # Add .0 as release version if not provided.
                if not match.group(2):
                    master_version = "{0}.0".format(match.group(1))
                else:
                    master_version = match.group(1)
            master_id = '{0}@{1}'.format(master.host, master.port)
            # Store the master version.
            srv_versions[master_version] = [master_id]

        # Get the server version for all slaves.
        for slave_key in self._slaves:
            slave = self._get_slave(slave_key)
            version = slave.get_version()
            match = re.match(_RE_VERSION_FORMAT, version.strip())
            if match:
                # Add .0 as release version if not provided.
                if not match.group(2):
                    version = "{0}.0".format(match.group(1))
                else:
                    version = match.group(1)
            # Store the slave version.
            if version in srv_versions:
                srv_versions[version].append(slave_key)
            else:
                srv_versions[version] = [slave_key]

        # Check the servers versions and issue a warning if different.
        if len(srv_versions) > 1:
            print("# WARNING: Servers using different versions:")
            for version in srv_versions:
                servers_str = ",".join(srv_versions[version])
                print("# - {0} for {1}.".format(version, servers_str))
            print("#")

    def check_gtid_sync(self):
        """Check GTIDs synchronization.

        Perform several GTID checks (enabled and errant transactions). If the
        master is available (was specified) then it also checks if GTIDs are
        in sync between master and its slaves and report the amount of
        transaction (i.e., GTIDs) behind the master for each slave.

        GTID differences might be an indicator of the existence of data
        consistency issues.

        Note: The master may not be specified, its use is not mandatory.
        """
        # Check if GTIDs are enabled on the topology.
        if self._get_master():  # Use of Master is not mandatory.
            # GTIDs must be enabled on the master.
            if self._get_master().supports_gtid().upper() != 'ON':
                raise UtilError(
                    "Master must support GTIDs and have GTID_MODE=ON."
                )
        # Skip slaves without GTID enabled and warn user.
        reset_base_srv = False
        for slave_key, slave_dict in self._slaves.items():
            slave = slave_dict['instance']
            support_gtid = slave.supports_gtid().upper()
            if support_gtid != 'ON':
                reason = "GTID_MODE=OFF" if support_gtid == 'OFF' \
                    else "not support GTIDs"
                print("# WARNING: Slave '{0}' will be skipped - "
                      "{1}.".format(slave_key, reason))
                print("#")
                del self._slaves[slave_key]
                self._rpl_topology.remove_slave(slave_dict)
                if slave_key == self._base_server_key:
                    reset_base_srv = True
        # At least on slave must have GTIDs enabled.
        if len(self._slaves) == 0:
            raise UtilError("No slaves found with GTID support and "
                            "GTID_MODE=ON.")
        # Reset base server if needed (it must have GTID_MODE=ON).
        if reset_base_srv:
            self._set_base_server()

        # Check the set of executed GTIDs and report differences, only if the
        # master is specified.
        if self._get_master():
            master_gtids = self._get_master().get_gtid_executed()
            slaves_gtids_data = \
                self._rpl_topology.slaves_gtid_subtract_executed(
                    master_gtids, multithreading=True
                )
            print("#\n# GTID differences between Master and Slaves:")
            for host, port, gtids_missing in slaves_gtids_data:
                slave_key = '{0}@{1}'.format(host, port)
                gtid_size = gtid_set_cardinality(gtids_missing)
                if gtid_size:
                    plural = 's' if gtid_size > 1 else ''
                    print("# - Slave '{0}' is {1} transaction{2} behind "
                          "Master.".format(slave_key, gtid_size, plural))
                    if self._verbosity:
                        print("#       Missing GTIDs: "
                              "{0}".format(gtids_missing))
                else:
                    print("# - Slave '{0}' is up-to-date.".format(slave_key))

        print("#")

    @staticmethod
    def _exist_in_obj_list(obj_name, obj_type, obj_list):
        """Check if object (name and type) exists in the given list.

        This function checks if the database object for the specified name and
        type exists in the specified list of database objects.

        obj_name[in]    Name of the object to check.
        obj_type[in]    Type of the object to check.
        obj_list[in]    List of objects to check. It is assumed that the list
                        has the format of the ones returned by the function
                        mysql.utilities.command.dbcompare.get_common_objects().
                        More precisely with the format:
                        [(obj_type1, (obj_name1,))..(obj_typeN, (obj_nameN,))]

        Returns a boolean value indicating if object with the specified name
        and type exists in the specified list of objects.
        """
        for obj_row in obj_list:
            if obj_row[0] == obj_type and obj_row[1][0] == obj_name:
                return True
        return False

    def _split_active_slaves(self, slaves):
        """Get the list of slaves with replication running and not.

        This method separates the list of given slaves into active (with the
        IO and SQL thread running) and non active slaves (with one of the
        threads stopped).

        slaves[in]      List of target slaves to separate.

        Returns a tuple with two elements, first with the list of active slaves
        and the second with the list of not active ones.
        """
        # Get slaves status.
        slaves_state = self._apply_for_all_slaves(slaves, 'get_slaves_errors',
                                                  multithreading=True)

        # Store IO and SQL thread status.
        active_slaves = []
        not_active_slaves = []
        for slave_key, state in slaves_state:
            # Locally store IO and SQL threads status.
            io_running = state[3].upper() == 'YES'
            self._slaves[slave_key]['IO_Running'] = io_running
            sql_running = state[4].upper() == 'YES'
            self._slaves[slave_key]['SQL_Running'] = sql_running
            if io_running and sql_running:
                active_slaves.append(slave_key)
            else:
                not_active_slaves.append(slave_key)
                print("#   WARNING: Slave not active '{0}' - "
                      "Sync skipped.".format(slave_key))
                if self._verbosity:
                    # Print warning if slave is stopped due to an error.
                    if not io_running and state[2]:
                        print("#    - IO thread stopped: ERROR {0} - "
                              "{1}".format(state[1], state[2]))
                    if not sql_running and state[6]:
                        print("#    - SQL thread stopped: ERROR {0} - "
                              "{1}".format(state[5], state[6]))

        # Return separated list of active and non active replication slaves.
        return active_slaves, not_active_slaves

    def _compute_sync_point(self, active_slaves=None):
        """Compute the GTID synchronization point.

        This method computes the GTID synchronization point based based on the
        GTID_EXECUTED set. If a master is available for synchronization the
        last GTID from the GTID_EXECUTED set is used as sync point  If no
        master is available the union of the GTID_EXECUTED sets among all
        active slaves is used as the sync point.

        active_slaves[in]   List of active slaves to consider. Only required
                            if the master is not available. It is assumed
                            that the list is composed by strings with the
                            format 'host@port', identifying each slave.

        Return a GTID set representing to synchronization point (to wait for
        slaves to catch up and stop).
        """
        if self._get_master():
            gtid_set = self._get_master().get_gtid_executed()
            master_uuid = self._get_master().get_server_uuid()
            return get_last_server_gtid(gtid_set, master_uuid)
        else:
            # Get GTID_EXECUTED on all slaves.
            all_gtid_executed = self._apply_for_all_slaves(
                active_slaves, 'get_gtid_executed', multithreading=True
            )

            # Compute the union of all GTID sets for each UUID among slaves.
            gtid_sets_by_uuid = {}
            for _, gtid_executed in all_gtid_executed:
                gtids_list = gtid_executed.split("\n")
                for gtid in gtids_list:
                    gtid_set = gtid.rstrip(', ')
                    uuid = gtid_set.split(':')[0]
                    if uuid not in gtid_sets_by_uuid:
                        gtid_sets_by_uuid[uuid] = gtid_set
                    else:
                        union_set = gtid_set_union(gtid_sets_by_uuid[uuid],
                                                   gtid_set)
                        gtid_sets_by_uuid[uuid] = union_set

            # Return union of all know executed GTID.
            return ",".join(gtid_sets_by_uuid.itervalues())

    def _sync_slaves(self, slaves, gtid):
        """Set synchronization point (specified GTID set) for the given slaves.

        The method set the synchronization point for the given slaves by
        (concurrently) stopping and immediately executing START SLAVE UNTIL
        on all given slaves in order to stop upon reaching the given GTID set
        (i.e., committing all corresponding transactions for the given GTID
        sync point).

        slaves[in]      List of target slaves to synchronize (i.e., instruct
                        to stop upon reaching the synchronization point).
        gtid[in]        GTID set used as the synchronization point.
        """
        # Make running slaves stop until sync point (GTID) is reached.
        if self._verbosity:
            print("#   Setting data synchronization point for slaves.")
        # STOP slave (only SQL thread).
        self._apply_for_all_slaves(slaves, 'stop_sql_thread',
                                   multithreading=True)
        # START slave UNTIL sync point is reached.
        # Note: Only the SQL thread is stopped when the condition is reached.
        until_ops = {'until_gtid_set': gtid, 'sql_after_gtid': True,
                     'only_sql_thread': True}
        self._apply_for_all_slaves(slaves, 'start', (), until_ops,
                                   multithreading=True)

    def _checksum_and_resume_rpl(self, not_sync_slaves, sync_slave, table):
        """Checksum table and resume replication on slaves.

        This method computes (concurrently) the table checksum of the given
        slaves lists (those synced and not synced). For the list of not synced
        slaves the table checksum is immediately computed. For the list of
        synced slaves, first it waits for them to catch up and the sync point
        and only then compute the table checksum and resume replication.

        not_sync_slaves[in] List of not synced slaves.
        sync_slave[in]      List of (previously) synced slaves.
        table[in]           Target table to compute the checksum.

        Returns a list of tuples, each tuple containing the identification of
        the server and the corresponding checksum result.
        """
        if self._verbosity:
            print("#   Compute checksum on slaves (wait to catch up and resume"
                  " replication).")
            sys.stdout.flush()
        not_sync_checksum = []
        if not_sync_slaves:
            not_sync_checksum = self._apply_for_all_slaves(
                not_sync_slaves, 'checksum_table', (table,),
                {'exec_timeout': self._checksum_timeout},
                multithreading=True
            )
        sync_checksum = []
        if sync_slave:
            sync_checksum = self._apply_for_all_slaves(
                sync_slave, 'wait_checksum_and_start', (table,),
                {'wait_timeout': self._rpl_timeout,
                 'wait_interval': self._interval,
                 'checksum_timeout': self._checksum_timeout},
                multithreading=True
            )
        return not_sync_checksum + sync_checksum

    def _check_table_data_sync(self, table, slaves):
        """Check table data synchronization for specified slaves.

        This method check the data consistency for the specified table between
        the base server (master or slave) and the specified salves. This
        operation requires the definition of a "synchronization point" in order
        to ensure that the "supposed" same data is compared between servers.
        This coordination process is based on GTIDs (checking that all data
        until a given GTID has been processed on the slaves). A different
        algorithm is used to set the "synchronization point" depending if the
        master is used or not. The data consistency is checked relying on the
        CHECKSUM TABLE query.

        If an error occur during this process, any locked table must be
        unlocked and both master and slaves should resume their previous
        activity.

        Important note: this method assumes that the table exists on the base
        server and all specified slaves, therefore checking the existence of
        the table as well as other integrity checks (server versions, GTID
        definitions, etc.) need to be performed outside the scope of this
        method.

        table[in]       Qualified name of the table to check (quoted with
                        backticks).
        slaves[in]      List of slaves to check. Each element of the list must
                        be a string with the format 'host@port'.

        Returns the number of data consistency found.
        """
        success = False
        checksum_issues = 0

        # If no master used then add base server (slave) to slaves to sync.
        if not self._get_master():
            slaves = slaves + [self._base_server_key]

        # Separate active from non active slaves.
        active_slaves, not_active_slaves = self._split_active_slaves(slaves)

        if self._get_master():
            # Lock the table on the master to get GTID synchronization point
            # and perform the table checksum.
            try:
                self._get_master().exec_query(
                    "LOCK TABLES {0} READ".format(table)
                )

                last_exec_gtid = self._compute_sync_point()
                if self._verbosity > 2:
                    print("#   Sync point GTID: {0}".format(last_exec_gtid))

                # Immediately instruct active slaves to stop on sync point.
                if active_slaves:
                    self._sync_slaves(active_slaves, last_exec_gtid)

                # Perform table checksum on master.
                base_server_checksum = self._get_master().checksum_table(
                    table, self._checksum_timeout
                )
                if base_server_checksum[0]:
                    success = True  # Successful checksum for base server.
                    if self._verbosity > 2:
                        print("#   Checksum on base server (Master): "
                              "{0}".format(base_server_checksum[0][1]))
                else:
                    print("#   [SKIP] {0} checksum on base server (Master) - "
                          "{1}".format(table, base_server_checksum[1]))
            finally:
                # Unlock table.
                self._get_master().exec_query("UNLOCK TABLES")
        elif active_slaves:
            # Perform sync without master, only based on active slave (if any).
            try:
                # Stop all active slaves to get the GTID synchronization point.
                self._apply_for_all_slaves(
                    active_slaves, 'stop_sql_thread', multithreading=True
                )

                sync_gtids = self._compute_sync_point(active_slaves)
                if self._verbosity > 2:
                    print("#   Sync point GTID: {0}".format(sync_gtids))

                # Instruct active slaves to stop on sync point.
                self._sync_slaves(active_slaves, sync_gtids)

            except UtilError:
                # Try to restart the slaves in case an error occurs.
                self._apply_for_all_slaves(
                    active_slaves, 'star_sql_thread', multithreading=True
                )

        # Compute checksum on all slaves and return to previous state.
        slaves_checksum = self._checksum_and_resume_rpl(not_active_slaves,
                                                        active_slaves, table)

        # Check if checksum for base server was successfully computed.
        if not self._get_master():
            for slave_key, checksum in slaves_checksum:
                if slave_key == self._base_server_key:
                    if checksum[0]:
                        success = True  # Successful checksum for base server.
                        base_server_checksum = checksum
                        slaves_checksum.remove((slave_key, checksum))
                        if self._verbosity > 2:
                            print("#   Checksum on base server: "
                                  "{0}".format(base_server_checksum[0][1]))
                    else:
                        print("#   [SKIP] {0} checksum on base server - "
                              "{1}".format(table, checksum[1]))
                    break

        # Compare checksum and report results.
        if success and slaves_checksum:
            for slave_key, checksum_res in slaves_checksum:
                if checksum_res[0] is None:
                    print("#   [SKIP] {0} checksum for Slave '{1}' - "
                          "{2}.".format(table, slave_key, checksum_res[1]))
                else:
                    if self._verbosity > 2:
                        checksum_val = ': {0}'.format(checksum_res[0][1])
                    else:
                        checksum_val = ''
                    if checksum_res[0] != base_server_checksum[0]:
                        print("#   [DIFF] {0} checksum for server '{1}'"
                              "{2}.".format(table, slave_key, checksum_val))
                        checksum_issues += 1
                    else:
                        print("#   [OK] {0} checksum for server '{1}'"
                              "{2}.".format(table, slave_key, checksum_val))

        return checksum_issues

    def check_data_sync(self, options, data_to_include, data_to_exclude):
        """Check data synchronization.

        Check if the data (in all tables) is in sync between the checked
        servers (master and its slaves, or only slaves). It reports structure
        difference database/tables missing or with a different definition and
        data differences between a base server and the others.

        Note: A different algorithm is applied to perform the synchronization,
        depending if the master is specified (available) or not.

        options[in]         Dictionary of options.
        data_to_include[in] Dictionary of data (set of tables) by database to
                            check.
        data_to_exclude[in] Dictionary of data (set of tables) by database to
                            exclude from check.

        Returns the number of consistency issues found (comparing database
        definitions and data).
        """
        issues_count = 0

        # Skip all database objects, except tables.
        options['skip_views'] = True
        options['skip_triggers'] = True
        options['skip_procs'] = True
        options['skip_funcs'] = True
        options['skip_events'] = True
        options['skip_grants'] = True

        diff_options = {}
        diff_options.update(options)
        diff_options['quiet'] = True  # Do not print messages.
        diff_options['suppress_sql'] = True  # Do not print SQL statements.
        diff_options['skip_table_opts'] = True  # Ignore AUTO_INCREMENT diffs.

        # Check the server version requirement to support sync features.
        # Slave servers of version >= 5.6.14 are required due to a known issue
        # for START SLAVE UNTIL with the SQL_AFTER_GTIDS option. More info:
        # https://dev.mysql.com/doc/refman/5.6/en/start-slave.html
        for slave_key in self._slaves:
            if not self._get_slave(slave_key).check_version_compat(5, 6, 14):
                raise UtilError(
                    "Server '{0}' version must be 5.6.14 or greater. Sync is "
                    "not supported for versions prior to 5.6.14 due to a "
                    "known issue with START SLAVE UNTIL and the "
                    "SQL_AFTER_GTIDS option.".format(slave_key))

        print("# Checking data consistency.\n#")
        base_srv_type = 'Master' if self._get_master() else 'Slave'
        print("# Using {0} '{1}' as base server for comparison."
              "".format(base_srv_type, self._base_server_key))

        # Get all databases from the base server.
        db_rows = self._base_server.get_all_databases()
        base_server_dbs = set([row[0] for row in db_rows])

        # Process databases to include/exclude from check.
        db_to_include = set()
        if data_to_include:
            db_to_include = set([db for db in data_to_include])
            base_server_dbs = base_server_dbs & db_to_include
            not_exist_db = db_to_include - base_server_dbs
            if not_exist_db:
                plurals = ('s', '') if len(not_exist_db) > 1 else ('', 'es')
                print('# WARNING: specified database{0} to check do{1} not '
                      'exist on base server and will be skipped: '
                      '{2}.'.format(plurals[0], plurals[1],
                                    ", ".join(not_exist_db)))
        db_to_exclude = set()
        if data_to_exclude:
            db_to_exclude = set(
                [db for db in data_to_exclude if not data_to_exclude[db]]
            )
            base_server_dbs = base_server_dbs - db_to_exclude

        # Check databases on slaves (except the base server).
        slaves_except_base = [key for key in self._slaves
                              if key != self._base_server_key]
        for slave_key in slaves_except_base:
            slave = self._get_slave(slave_key)
            db_rows = slave.get_all_databases()
            slave_dbs = set([row[0] for row in db_rows])
            # Process databases to include/exclude.
            if db_to_include:
                slave_dbs = slave_dbs & db_to_include
            if db_to_exclude:
                slave_dbs = slave_dbs - db_to_exclude
            # Add slave databases set to internal state.
            self._slaves[slave_key]['databases'] = slave_dbs
            # Report databases not on base server and filtered by replication.
            dbs_not_in_base_srv = slave_dbs - base_server_dbs
            filtered_dbs = set(
                [db for db in dbs_not_in_base_srv
                    if self._is_rpl_filtered(db, slave=self._base_server_key)]
            )
            dbs_not_in_base_srv -= filtered_dbs
            for db in filtered_dbs:
                print("# [SKIP] Database '{0}' - filtered by replication "
                      "rule on base server.".format(db))
            if dbs_not_in_base_srv:
                issues_count += len(dbs_not_in_base_srv)
                plural = 's' if len(dbs_not_in_base_srv) > 1 else ''
                print("# [DIFF] Database{0} NOT on base server but found on "
                      "'{1}': {2}".format(plural, slave_key,
                                          ",".join(dbs_not_in_base_srv)))

        # Determine server to check base replication filtering options.
        filter_srv = None if self._get_master() else self._base_server_key

        # Check data consistency for each table on the base server.
        for db_name in base_server_dbs:
            # Skip database if filtered by defined replication rules.
            if self._is_rpl_filtered(db_name, slave=filter_srv):
                print("# [SKIP] Database '{0}' check - filtered by "
                      "replication rule.".format(db_name))
                continue
            print("# Checking '{0}' database...".format(db_name))
            slaves_to_check = {}
            # Check if database exists on slaves (except the base server).
            for slave_key in slaves_except_base:
                # Skip database if filtered by defined replication rules.
                if self._is_rpl_filtered(db_name, slave=slave_key):
                    print("# [SKIP] Database '{0}' check for '{1}' - filtered "
                          "by replication rule.".format(db_name, slave_key))
                    continue
                if db_name in self._slaves[slave_key]['databases']:
                    # Store slave database instance and common objects.
                    slave_db = Database(self._get_slave(slave_key), db_name,
                                        options)
                    slave_db.init()
                    slave_dic = {'db': slave_db}
                    in_both, in_basesrv, not_in_basesrv = get_common_objects(
                        self._base_server, self._get_slave(slave_key),
                        db_name, db_name, False, options)
                    # Process tables to include/exclude from check (on slaves).
                    if (data_to_include and db_name in data_to_include
                            and data_to_include[db_name]):
                        in_both = [
                            obj_row for obj_row in in_both
                            if obj_row[1][0] in data_to_include[db_name]
                        ]
                        in_basesrv = [
                            obj_row for obj_row in in_basesrv
                            if obj_row[1][0] in data_to_include[db_name]
                        ]
                        not_in_basesrv = [
                            obj_row for obj_row in not_in_basesrv
                            if obj_row[1][0] in data_to_include[db_name]
                        ]
                    if (data_to_exclude and db_name in data_to_exclude
                            and data_to_exclude[db_name]):
                        in_both = [
                            obj_row for obj_row in in_both
                            if obj_row[1][0] not in data_to_exclude[db_name]
                        ]
                        in_basesrv = [
                            obj_row for obj_row in in_basesrv
                            if obj_row[1][0] not in data_to_exclude[db_name]
                        ]
                        not_in_basesrv = [
                            obj_row for obj_row in not_in_basesrv
                            if obj_row[1][0] not in data_to_exclude[db_name]
                        ]
                    slave_dic['in_both'] = in_both
                    slave_dic['in_basesrv'] = in_basesrv
                    slaves_to_check[slave_key] = slave_dic
                    # Report tables not on base server and filtered by
                    # replication.
                    tbls_not_in = set(
                        [obj_row[1][0] for obj_row in not_in_basesrv
                         if obj_row[0] == 'TABLE']
                    )
                    filtered_tbls = set(
                        [tbl for tbl in tbls_not_in if self._is_rpl_filtered(
                            db_name, tbl_name=tbl, slave=self._base_server_key
                        )]
                    )
                    tbls_not_in -= filtered_tbls
                    for tbl in filtered_tbls:
                        print("# [SKIP] Table '{0}' - filtered by replication "
                              "rule on base server.".format(tbl))
                    if tbls_not_in:
                        plural = 's' if len(tbls_not_in) > 1 else ''
                        print("#   [DIFF] Table{0} NOT on base server but "
                              "found on '{1}': "
                              "{2}".format(plural, slave_key,
                                           ", ".join(tbls_not_in)))
                        issues_count += len(tbls_not_in)
                else:
                    print("#   [DIFF] Database '{0}' NOT on server "
                          "'{1}'.".format(db_name, slave_key))
                    issues_count += 1
            # Only check database if at least one slave has it.
            if slaves_to_check:
                db = Database(self._base_server, db_name, options)
                db.init()
                for db_obj in db.get_next_object():
                    obj_type = db_obj[0]
                    obj_name = db_obj[1][0]
                    # Process tables to include/exclude from check (on base
                    # server).
                    if (data_to_include and data_to_include[db_name]
                            and obj_name not in data_to_include[db_name]):
                        # Skip to the next object if not in data to include.
                        continue
                    if (data_to_exclude and data_to_exclude[db_name]
                            and obj_name in data_to_exclude[db_name]):
                        # Skip to the next object if in data to exclude.
                        continue
                    checksum_task = []
                    # Check object data on all valid slaves.
                    for slave_key in slaves_to_check:
                        # Skip table if filtered by defined replication rules.
                        if (obj_type == 'TABLE'
                            and self._is_rpl_filtered(db_name, obj_name,
                                                      slave=slave_key)):
                            print("# [SKIP] Table '{0}' check for '{1}' - "
                                  "filtered by replication rule."
                                  "".format(obj_name, slave_key))
                            continue
                        slave_dic = slaves_to_check[slave_key]
                        # Check if object does not exist on Slave.
                        if self._exist_in_obj_list(obj_name, obj_type,
                                                   slave_dic['in_basesrv']):
                            print("#   [DIFF] {0} '{1}.{2}' NOT on server "
                                  "'{3}'.".format(obj_type.capitalize(),
                                                  db_name, obj_name,
                                                  slave_key))
                            issues_count += 1
                            continue

                        # Quote object name with backticks.
                        q_obj = '{0}.{1}'.format(
                            quote_with_backticks(db_name),
                            quote_with_backticks(obj_name)
                        )

                        # Check object definition.
                        def_diff = diff_objects(
                            self._base_server, self._get_slave(slave_key),
                            q_obj, q_obj, diff_options, obj_type
                        )
                        if def_diff:
                            print("#   [DIFF] {0} {1} definition is "
                                  "different on '{2}'."
                                  "".format(obj_type.capitalize(), q_obj,
                                            slave_key))
                            issues_count += 1
                            if self._verbosity:
                                for diff in def_diff[3:]:
                                    print("#       {0}".format(diff))
                            continue

                        # Add slave to table checksum task.
                        checksum_task.append(slave_key)

                    # Perform table checksum on valid slaves.
                    if checksum_task and obj_type == 'TABLE':
                        print("# - Checking '{0}' table data..."
                              "".format(obj_name))
                        num_issues = self._check_table_data_sync(q_obj,
                                                                 checksum_task)
                        issues_count += num_issues

        print("#\n#...done.\n#")
        str_issues_count = 'No' if issues_count == 0 else str(issues_count)
        plural = 's' if issues_count > 1 else ''
        print("# SUMMARY: {0} data consistency issue{1} found.\n"
              "#".format(str_issues_count, plural))
        return issues_count



#!/usr/bin/env python
#
# Copyright (c) 2011, 2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
This file contains the replicate utility. It is used to establish a
master/slave replication topology among two servers.
"""

from mysql.utilities.common.tools import check_python_version

# Check Python version compatibility
check_python_version()

import os.path
import sys

from mysql.utilities.exception import UtilError, FormatError
from mysql.utilities.command.check_rpl import check_replication
from mysql.utilities.common.options import add_verbosity, setup_common_options
from mysql.utilities.common.ip_parser import parse_connection
from mysql.utilities.common.server import check_hostname_alias
from mysql.utilities.common.tools import check_connector_python
from mysql.utilities.common.messages import (PARSE_ERR_OPTS_REQ,
                                             WARN_OPT_USING_DEFAULT)


# Constants
NAME = "MySQL Utilities - mysqlrplcheck "
DESCRIPTION = "mysqlrplcheck - check replication"
USAGE = "%prog --master=root@localhost:3306 --slave=root@localhost:3310 "

PRINT_WIDTH = 75

# Check for connector/python
if not check_connector_python():
    sys.exit(1)

if __name__ == '__main__':
    # Setup the command parser
    parser = setup_common_options(os.path.basename(sys.argv[0]),
                                  DESCRIPTION, USAGE, True, False)

    # Setup utility-specific options:

    # Connection information for the source server
    parser.add_option('--master', action="store", dest="master",
                      type="string", default=None,
                      help="connection information for master server in the "
                           "form: <user>[:<password>]@<host>[:<port>]"
                           "[:<socket>] or <login-path>[:<port>][:<socket>].")

    # Connection information for the destination server
    parser.add_option('--slave', action="store", dest="slave",
                      type="string", default=None,
                      help="connection information for slave server in the "
                           "form: <user>[:<password>]@<host>[:<port>]"
                           "[:<socket>] or <login-path>[:<port>][:<socket>]")

    # Add --master-info-file
    parser.add_option("--master-info-file", action="store", dest="master_info",
                      type="string", default="master.info",
                      help="the name of the master information file on the "
                           "slave. Default = 'master.info' read from the data "
                           "directory. Note: this option requires that the "
                           "utility run on the slave with appropriate file "
                           "read access to the data directory.")

    # Add --show-slave-status
    parser.add_option("--show-slave-status", "-s", action="store_true",
                      dest="slave_status", default=False,
                      help="show slave status")

    # Add display width option
    parser.add_option("--width", action="store", dest="width",
                      type="int", help="display width",
                      default=PRINT_WIDTH)

    # Add suppress to suppress warning messages
    parser.add_option("--suppress", action="store_true", dest="suppress",
                      default=False, help="suppress warning messages")

    # Add verbosity
    add_verbosity(parser, True)

    # Now we process the rest of the arguments.
    opt, args = parser.parse_args()

    # option --master is required (mandatory)
    if not opt.master:
        default_val = 'root@localhost:3306'
        print(WARN_OPT_USING_DEFAULT.format(default=default_val,
                                            opt='--master'))
        # Print the WARNING to force determinism if a parser error occurs.
        sys.stdout.flush()

    # option --slave is required (mandatory)
    if not opt.slave:
        parser.error(PARSE_ERR_OPTS_REQ.format(opt='--slave'))

    # Parse source connection values
    try:
        m_values = parse_connection(opt.master)
    except FormatError:
        _, err, _ = sys.exc_info()
        parser.error("Master connection values invalid: %s." % err)
    except UtilError:
        _, err, _ = sys.exc_info()
        parser.error("Master connection values invalid: %s." % err.errmsg)

    # Parse source connection values
    try:
        s_values = parse_connection(opt.slave)
    except FormatError:
        _, err, _ = sys.exc_info()
        parser.error("Slave connection values invalid: %s." % err)
    except UtilError:
        _, err, _ = sys.exc_info()
        parser.error("Slave connection values invalid: %s." % err.errmsg)

    # Check hostname alias
    if check_hostname_alias(m_values, s_values):
        parser.error("The master and slave are the same host and port.")

    # Create dictionary of options
    options = {
        'verbosity': opt.verbosity,
        'pedantic': False,
        'quiet': opt.quiet,
        'suppress': opt.suppress,
        'master_info': opt.master_info,
        'slave_status': opt.slave_status,
        'width': opt.width
    }

    try:
        res = check_replication(m_values, s_values, options)
        if res:
            sys.exit(1)
    except UtilError:
        _, e, _ = sys.exc_info()
        print("ERROR: {0}".format(e.errmsg))
        sys.exit(1)

    sys.exit()



#!/usr/bin/env python
#
# Copyright (c) 2010, 2013, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
This file contains the export database utility which allows users to export
metadata for objects in a database and data for tables.
"""

from mysql.utilities.common.tools import check_python_version

# Check Python version compatibility
check_python_version()

import multiprocessing
import os
import shutil
import sys
import tempfile
import time

from mysql.utilities.command.dbexport import export_databases
from mysql.utilities.command.dbexport import multiprocess_db_export_task
from mysql.utilities.common.ip_parser import parse_connection
from mysql.utilities.common.options import (
    add_all, add_character_set_option, add_format_option, add_locking,
    add_regexp, add_rpl_mode, add_rpl_user, add_skip_options, add_verbosity,
    check_all, check_rpl_options, check_skip_options, check_verbosity,
    setup_common_options
)
from mysql.utilities.common.sql_transform import (is_quoted_with_backticks,
                                                  remove_backtick_quoting)
from mysql.utilities.common.tools import (check_connector_python,
                                          print_elapsed_time)

from mysql.utilities.exception import FormatError
from mysql.utilities.exception import UtilError

# Constants
NAME = "MySQL Utilities - mysqldbexport "
DESCRIPTION = "mysqldbexport - export metadata and data from databases"
USAGE = "%prog --server=user:pass@host:port:socket db1, db2, db3"

_PERMITTED_DISPLAY = ["names", "brief", "full"]
_PERMITTED_EXPORTS = ["data", "definitions", "both"]

# Check for connector/python
if not check_connector_python():
    sys.exit(1)

if __name__ == '__main__':
    # Needed for freeze support to avoid RuntimeError when running as a Windows
    # executable, otherwise ignored.
    multiprocessing.freeze_support()

    # Setup the command parser and setup server, help
    parser = setup_common_options(os.path.basename(sys.argv[0]),
                                  DESCRIPTION, USAGE)

    # Setup utility-specific options:

    # Add character set option
    add_character_set_option(parser)

    # Output format
    add_format_option(parser, "display the output in either sql (default), "
                      "grid, tab, csv, or vertical format", "sql", True)

    # Display format
    parser.add_option("-d", "--display", action="store", dest="display",
                      default="brief", help="control the number of columns "
                      "shown: 'brief' = minimal columns for object creation "
                      "(default), 'full' = all columns, 'names' = only object "
                      "names (not valid for --format=sql)", type="choice",
                      choices=_PERMITTED_DISPLAY)

    # Export mode
    parser.add_option("-e", "--export", action="store", dest="export",
                      default="definitions", help="control the export of "
                      "either 'data' = only the table data for the tables in "
                      "the database list, 'definitions' = export only the "
                      "definitions for the objects in the database list, or "
                      "'both' = export the metadata followed by the data "
                      "(default: export definitions)", type="choice",
                      choices=_PERMITTED_EXPORTS)

    # Single insert mode
    parser.add_option("-b", "--bulk-insert", action="store_true",
                      dest="bulk_import", default=False,
                      help="use bulk insert statements for data "
                           "(default:False)")

    # Header row
    parser.add_option("-h", "--no-headers", action="store_true",
                      dest="no_headers", default=False, help="do not display "
                      "the column headers - ignored for grid format")

    # Skip blobs for export
    parser.add_option("--skip-blobs", action="store_true", dest="skip_blobs",
                      default=False, help="do not export blob data.")

    # File-per-table mode
    parser.add_option("--file-per-table", action="store_true",
                      dest="file_per_tbl", default=False, help="write table "
                      "data to separate files. Valid only for --export=data "
                      "or --export=both.")

    # Add the exclude database option
    parser.add_option("-x", "--exclude", action="append", dest="exclude",
                      type="string", default=None, help="exclude one or more "
                      "objects from the operation using either a specific "
                      "name (e.g. db1.t1), a LIKE pattern (e.g. db1.t% or "
                      "db%.%) or a REGEXP search pattern. To use a REGEXP "
                      "search pattern for all exclusions, you must also "
                      "specify the --regexp option. Repeat the --exclude "
                      "option for multiple exclusions.")

    # Add the all database options
    add_all(parser, "databases")

    # Add the skip common options
    add_skip_options(parser)

    # Add verbosity and quiet (silent) mode
    add_verbosity(parser, True)

    # Add regexp
    add_regexp(parser)

    # Add locking
    add_locking(parser)

    # Replication user and password
    add_rpl_user(parser, None)

    # Add replication options
    add_rpl_mode(parser)

    parser.add_option("--skip-gtid", action="store_true", default=False,
                      dest="skip_gtid", help="skip creation of GTID_PURGED "
                      "statements.")

    # Add comment replication output
    parser.add_option("--comment-rpl", action="store_true", default=False,
                      dest="comment_rpl", help="place the replication "
                      "statements in comment statements. Valid only with "
                      "--rpl option.")

    parser.add_option("--skip-fkey-checks", action="store_true", default=False,
                      dest="skip_fkeys", help="skip creation of foreign key "
                      "disable/enable statements.")

    # Add multiprocessing option.
    parser.add_option("--multiprocess", action="store", dest="multiprocess",
                      type="int", default="1", help="use multiprocessing, "
                      "number of processes to use for concurrent execution. "
                      "Special values: 0 (number of processes equal to the "
                      "CPUs detected) and 1 (default - no concurrency).")

    # Add output file option.
    parser.add_option("--output-file", action="store", dest="output_file",
                      help="path and file name to store the generated output, "
                           "by default the standard output (no file).")

    # Now we process the rest of the arguments.
    opt, args = parser.parse_args()

    # Warn if quiet and verbosity are both specified
    check_verbosity(opt)

    try:
        skips = check_skip_options(opt.skip_objects)
    except UtilError:
        _, err, _ = sys.exc_info()
        print("ERROR: {0}".format(err.errmsg))
        sys.exit(1)

    # Fail if no db arguments or all
    if len(args) == 0 and not opt.all:
        parser.error("You must specify at least one database to export or "
                     "use the --all option to export all databases.")

    # Check replication options
    check_rpl_options(parser, opt)

    # Fail if we have arguments and all databases option listed.
    check_all(parser, opt, args, "databases")

    if opt.skip_blobs and not opt.export == "data":
        print("# WARNING: --skip-blobs option ignored for metadata export.")

    if opt.file_per_tbl and opt.export in ("definitions", "both"):
        print("# WARNING: --file-per-table option ignored for metadata "
              "export.")

    if "data" in skips and opt.export == "data":
        print("ERROR: You cannot use --export=data and --skip-data when "
              "exporting table data.")
        sys.exit(1)

    # Process --exclude values to remove unnecessary quotes (when used) in
    # order to avoid further matching issues.
    if opt.exclude:
        # Remove unnecessary outer quotes.
        exclude_list = [pattern.strip("'\"") for pattern in opt.exclude]
    else:
        exclude_list = opt.exclude

    # Check multiprocessing options.
    if opt.multiprocess < 0:
        parser.error("Number of processes '{0}' must be greater or equal than "
                     "zero.".format(opt.multiprocess))
    num_cpu = multiprocessing.cpu_count()
    if opt.multiprocess > num_cpu and not opt.quiet:
        print("# WARNING: Number of processes '{0}' is greater than the "
              "number of CPUs '{1}'.".format(opt.multiprocess, num_cpu))

    # Warning for non-posix (windows) systems if too many process are used.
    num_db = len(args)
    if (os.name != 'posix' and num_db and opt.multiprocess > num_db
            and not opt.quiet):
        print("# WARNING: Number of processes '{0}' is greater than the "
              "number of databases to export '{1}'.".format(opt.multiprocess,
                                                            num_db))

    # Check output_file option.
    if opt.output_file:
        # Check if file already exists.
        if os.path.exists(opt.output_file) and not opt.quiet:
            print("# WARNING: Specified output file already exists. The file "
                  "will be overwritten.")
        output_filename = opt.output_file
        try:
            output_file = open(output_filename, 'w')
        except IOError:
            parser.error("Unable to create file (check path and access "
                         "privileges): {0}".format(opt.output_file))
    else:
        # Always send output to a file for performance reasons (contents sent
        # at the end to the stdout).
        output_file = tempfile.NamedTemporaryFile(delete=False)
        output_filename = None

    # Set options for database operations.
    options = {
        "skip_tables": "tables" in skips,
        "skip_views": "views" in skips,
        "skip_triggers": "triggers" in skips,
        "skip_procs": "procedures" in skips,
        "skip_funcs": "functions" in skips,
        "skip_events": "events" in skips,
        "skip_grants": "grants" in skips,
        "skip_create": "create_db" in skips,
        "skip_data": "data" in skips,
        "skip_blobs": opt.skip_blobs,
        "skip_fkeys": opt.skip_fkeys,
        "format": opt.format,
        "no_headers": opt.no_headers,
        "display": opt.display,
        "single": not opt.bulk_import,
        "quiet": opt.quiet,
        "verbosity": opt.verbosity,
        "debug": opt.verbosity >= 3,
        "file_per_tbl": opt.file_per_tbl,
        "exclude_patterns": exclude_list,
        "all": opt.all,
        "use_regexp": opt.use_regexp,
        "locking": opt.locking,
        "rpl_user": opt.rpl_user,
        "rpl_mode": opt.rpl_mode,
        "rpl_file": opt.rpl_file,
        "comment_rpl": opt.comment_rpl,
        "export": opt.export,
        "skip_gtid": opt.skip_gtid,
        "charset": opt.charset,
        "multiprocess": num_cpu if opt.multiprocess == 0 else opt.multiprocess,
        "output_filename": output_filename,
    }

    # Parse server connection values
    try:
        server_values = parse_connection(opt.server, None, options)
    except FormatError:
        _, err, _ = sys.exc_info()
        parser.error("Server connection values invalid: {0}.".format(err))
    except UtilError:
        _, err, _ = sys.exc_info()
        parser.error("Server connection values invalid: "
                     "{0}.".format(err.errmsg))

    # Build list of databases to copy
    db_list = []
    for db in args:
        # Remove backtick quotes (handled later)
        db = remove_backtick_quoting(db) \
            if is_quoted_with_backticks(db) else db
        db_list.append(db)

    try:
        # record start time
        if opt.verbosity >= 3:
            start_export_time = time.time()

        # Export databases concurrently for non posix systems (windows).
        if options['multiprocess'] > 1 and os.name != 'posix':
            # Create export databases tasks.
            export_db_tasks = []
            for db in db_list:
                export_task = {
                    'srv_con': server_values,
                    'db_list': [db],
                    'options': options,
                }
                export_db_tasks.append(export_task)

            # Create process pool.
            workers_pool = multiprocessing.Pool(
                processes=options['multiprocess']
            )

            # Concurrently export databases.
            res = workers_pool.map_async(multiprocess_db_export_task,
                                         export_db_tasks)
            workers_pool.close()
            # Get list of temporary files with the exported data.
            tmp_files_list = res.get()
            workers_pool.join()

            # Merge resulting temp files (if generated).
            for tmp_filename in tmp_files_list:
                if tmp_filename:
                    tmp_file = open(tmp_filename, 'r')
                    shutil.copyfileobj(tmp_file, output_file)
                    tmp_file.close()
                    os.remove(tmp_filename)
        else:
            # Export all specified databases (no database level concurrency).
            # Note: on POSIX systems multiprocessing is applied at the table
            # level (not database).
            export_databases(server_values, db_list, output_file, options)

        if output_filename is None:
            # Dump the export output to the stdout.
            output_file.seek(0)
            shutil.copyfileobj(output_file, sys.stdout)
            output_file.close()
            os.remove(output_file.name)

        # record elapsed time
        if opt.verbosity >= 3:
            sys.stdout.flush()
            print_elapsed_time(start_export_time)

    except UtilError:
        _, err, _ = sys.exc_info()
        print("ERROR: {0}".format(err.errmsg))
        sys.exit(1)

    sys.exit()



# MySQL Connector/Python - MySQL driver written in Python.


from django.db import models
from django.db.backends import BaseDatabaseValidation


class DatabaseValidation(BaseDatabaseValidation):
    def validate_field(self, errors, opts, f):
        """
        MySQL has the following field length restriction:
        No character (varchar) fields can have a length exceeding 255
        characters if they have a unique index on them.
        """
        varchar_fields = (models.CharField, models.CommaSeparatedIntegerField,
                          models.SlugField)
        if isinstance(f, varchar_fields) and f.max_length > 255 and f.unique:
            msg = ('"%(name)s": %(cls)s cannot have a "max_length" greater '
                   'than 255 when using "unique=True".')
            errors.add(opts, msg % {'name': f.name,
                                    'cls': f.__class__.__name__})



#
# Copyright (c) 2013,2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""This is the scheduler which is used to guarantee that conflicting procedures
cannot be concurrently executed. Locks are atomically acquired through the
LockManager class in one single step thus avoiding deadlock issues.
"""
import threading
import logging
import Queue

import mysql.fabric.errors as _errors
import mysql.fabric.utils as _utils

_LOGGER = logging.getLogger(__name__)

class Scheduler(object):
    """Class responsible for scheduling procedures.
    """
    def __init__(self):
        """Constructor for Scheduler.
        """
        self.__lock_manager = LockManager()
        self.__queue = Queue.Queue()

    @property
    def lock_manager(self):
        """Return a reference to the LockManager instance.
        """
        return self.__lock_manager

    def enqueue_procedures(self, procedures):
        """Enqueue a list of procedures.

        :param procedures: Reference to a list of procedures.
        """
        for procedure in procedures:
            self.enqueue_procedure(procedure)

    def enqueue_procedure(self, procedure):
        """Enqueue a procedure.

        :param procedure: Reference to a procedure.
        """
        if procedure:
            _LOGGER.debug("Enqueued procedure (%s).", procedure.uuid)
        self.__queue.put(procedure)

    def next_procedure(self, condition=None):
        """Get the next procedure to be executed.

        :param condition: Condition variable which is used to notify the
                          caller when the locks are acquired or another
                          thread breaks its locks.
        :return: Return a reference to a procedure that is ready to be
                 executed.
        :rtype: Procedure
        """
        procedure = self.__queue.get()
        if procedure is not None:
            _LOGGER.debug("Locking procedure (%s).", procedure.uuid)
            self.__lock_manager.lock(
                procedure,
                procedure.get_lockable_objects(),
                procedure.get_priority(),
                condition
            )
            _LOGGER.debug("Locked procedure (%s).", procedure.uuid)
        return procedure

    def done(self, procedure):
        """Inform that a procedure has been executed and unlock
        it.

        :param procedure: Reference to a procedure.
        """
        if procedure is not None:
            self.__lock_manager.release(procedure)
            self.__queue.task_done()
            _LOGGER.debug("Unlocked procedure (%s).", procedure.uuid)


class LockManager(object):
    """Class that implements a lock system.
    """
    def __init__(self):
        """Constructor for the LockSystem.
        """
        # Latch to prevent race conditions.
        self.__lock = threading.Condition()

        # Dictionary that works as a set of queues and maps objects to
        # threads that acquired their locks or are willing to acquire
        # them.
        self.__objects = {}

        # Dictionary that maps a procedure to a 3-tuple that contains
        # the objects that the procedure needs to lock, the thread's id
        # of the caller which acquired the locks and a condition
        # variable if there is any.
        self.__procedures = {}

        # List with procedures that acquire all the necessary locks and
        # can be executed.
        self.__free = []

    @property
    def objects(self):
        """Return a dictionary mapping all locked objects to the procedures
        that grabbed the lock and those that are waiting to do so::

            {obj_1 : [proc_1, proc_2], obj_2 : [proc_1], ...}

        :return: Dictionary with information on lock requests.
        :rtype: Dictionary
        """
        with self.__lock:
            return self.__objects.copy()

    @property
    def procedures(self):
        """Return a dictionary mapping all procedures to the objects that
        they are willing to lock, the threads that requested them and a
        lock object which can be used to notify the caller either when
        the locks are acquired or when they are broken by another thread::

            {proc_1 : ([obj_1, obj_2], thread_1, lock_1),
             proc_2 : ([obj_1], thread_2, lock_2),
             ...
            }

        :return: Dictionary with information on procedures.
        :rtype: Dictionary
        """
        with self.__lock:
            return self.__procedures.copy()

    @property
    def free(self):
        """Return a list with all procedures that can proceed with their
        execution::

            [proc_1]

        :return: A list with all procedures that can proceed with their
                 execution.
        :rtype: List
        """
        with self.__lock:
            return list(self.__free)

    def lock(self, procedure, objects, force=False, condition=None):
        """Request locks for a procedure and wait until they are acquired.
        If the force flag is set, all procedures that conflict with the
        current request will be aborted.

        :param procedure: Reference to procedure.
        :param objects: Set of objects that will be locked.
        :param force: Whether the request has priority over requests
                         previously processed or not.
        :param condition: Condition variable which is used to notify the
                          caller when the locks are acquired or another
                          thread breaks its locks.
        :return: Reference to the incomming procedure.
        """
        try:
            self.__lock.acquire()
            _LOGGER.debug(
                "LockManager - Enqueuing request for procedure (%s) "
                "on objects(%s).", procedure, objects
            )
            if force:
                self._break_conflicts(objects)
            self._enqueue(procedure, objects)
            self._set_notification(procedure, condition)
            return self._wait_for_procedure(procedure, condition)
        finally:
            try:
                self.__lock.release()
            except RuntimeError:
                pass

    def release(self, procedure):
        """Dequeue a procedure's request.

        :param procedure: Reference to procedure.
        """
        with self.__lock:
            _LOGGER.debug(
                "LockManager - Dequeuing request for procedure (%s).",
                procedure
            )
            self._dequeue(procedure)

    def enqueue(self, procedure, objects, force=False):
        """Enqueue a procedure's request. If the force flag is set,
        all procedures that conflict with the current request will be
        aborted.

        :param procedure: Reference to procedure.
        :param objects: Set of objects that will be locked.
        :param force: Whether the request has priority over requests
                      previously processed or not.
        """
        with self.__lock:
            _LOGGER.debug(
                "LockManager - Enqueuing request for procedure (%s) on "
                "objects(%s).", procedure, objects
            )
            if force:
                self._break_conflicts(objects)
            self._enqueue(procedure, objects)

    def get(self, procedure, condition=None):
        """Wait until the procedure acquires all the necessary locks.

        :param procedure: Reference to procedure.
        :param condition: Condition variable which is used to notify the
                          caller when the locks are acquired or another
                          thread breaks its locks.
        :return: Reference to the requested procedure.
        """
        try:
            self.__lock.acquire()
            self._procedure_enqueued(procedure)
            self._set_notification(procedure, condition)
            return self._wait_for_procedure(procedure, condition)
        finally:
            try:
                self.__lock.release()
            except RuntimeError:
                pass

    def break_conflicts(self, objects):
        """Make all procedures, which are holding or waiting for locks
        on a list of objects given as parameter, abort and raise an
        exception within the context of the thread that acquqired the
        locks.

        :param objects: Set of objects that may have associated requests.
        """
        with self.__lock:
            procedures = self._break_conflicts(objects)
            _LOGGER.debug(
                "The following procedures were aborted: (%s).", procedures
            )
            return procedures

    def check_conflicts(self, objects):
        """Return the set of procedures that are holding locks or waiting
        for locks on a list of objects given as parameter.

        :param objects: Set of objects that may have associated requests.
        """
        with self.__lock:
            procedures = self._check_conflicts(objects)
            _LOGGER.debug(
                "The procedures (%s) have locked or are willing to lock "
                "one of the objects in (%s).", procedures, objects
            )
            return procedures

    def _set_notification(self, procedure, condition):
        """Update information that can be used to notify threads on
        possible changes to the lock information.
        """
        assert(condition is None or isinstance(condition, threading.Condition))
        objects, _, _ = self.__procedures[procedure]
        thread = threading.current_thread()
        self.__procedures[procedure] = (objects, thread, condition)

    def _dequeue(self, procedure):
        """Dequeue a procedure's request.
        """
        # Dictionary that maps procedures to objects that it already
        # has acquired a lock on.
        head_procedures = set()
        objects, _, _ = self._procedure_enqueued(procedure)

        # Remove the information on the procedure from the procedures'
        # dictionary and from the free list if it is there.
        del self.__procedures[procedure]
        try:
            self.__free.remove(procedure)
        except ValueError:
            pass

        # Remove the information on the procedure from the objects'
        # dictionary.
        _LOGGER.debug("Released procedure - %s objects %s.", procedure, objects)
        for obj in objects:
            # The wait_queue contains the list of procedures willing
            # to lock the object.
            wait_queue = self.__objects[obj]
            wait_queue.remove(procedure)
            if len(wait_queue) == 0:
                # If the queue is empty, remove the object from the
                # dictionary.
                del self.__objects[obj]
            elif wait_queue[0] not in self.__free:
                # If a procedure is at the head of the queue it has a lock
                # on the object.
                head_procedures.add(wait_queue[0])

        _LOGGER.debug("Possible affected procedures %s.", head_procedures)
        for procedure in head_procedures:
            objects, _, _ = self.__procedures[procedure]
            procedures = set([self.__objects[obj][0] for obj in objects])
            assert(len(procedures) > 0)
            if procedures == set([procedure]):
                _LOGGER.debug("Procedure %s is ready to be executed.", procedure)
                # If the requested set of objects is equal to the acquired
                # set of objects, the procedure is ready to go.
                self.__free.append(procedure)
                self.__lock.notify_all()

    def _enqueue(self, procedure, objects):
        """Enqueue a procedure's request.
        """
        assert(isinstance(objects, set))

        try:
            # Verifying if a procedure is not already enqueued.
            objects, _, _ = self.__procedures[procedure]
            raise _errors.LockManagerError(
                "The procedure (%s) is already enqueued." % (procedure, )
            )
        except KeyError:
            pass

        ready = True
        for obj in objects:
            queue = self.__objects.get(obj, [])
            if len(queue) != 0:
                ready = False
            queue.append(procedure)
            self.__objects[obj] = queue

        self.__procedures[procedure] = (objects, None, None)

        if ready:
            self.__free.append(procedure)
            self.__lock.notify_all()

    def _check_conflicts(self, objects):
        """Return the set of procedures that are holding locks or waiting
        for locks on a list of objects given as parameter.
        """
        assert(isinstance(objects, set))
        procedures = [self.__objects.get(obj, []) for obj in objects]
        return list(set([proc for lst_proc in procedures for proc in lst_proc]))

    def _break_conflicts(self, objects):
        """Make all procedures, which are holding or waiting for locks
        on a list of objects given as parameter, abort and raise an
        exception within the context of the thread that acquqired the
        locks.
        """
        assert(isinstance(objects, set))
        procedures = self._check_conflicts(objects)
        for procedure in procedures:
            requested_objects, thread, condition = self.__procedures[procedure]
            assert(len(requested_objects & objects) != 0)
            self._dequeue(procedure)
            if thread and threading.current_thread() != thread:
                try:
                    _utils.async_raise(thread.ident, _errors.LockManagerError)
                except (ValueError, SystemError):
                    _LOGGER.debug(
                        "Error trying to notify thread (%s) that holds locks "
                        "for procedure (%s).", thread, procedure
                    )
            if condition:
                with condition:
                    condition.notify_all()
        return procedures

    def _wait_for_procedure(self, procedure, condition):
        """Wait until procedure acquires all locks requested.
        """
        not_free = True
        while not_free:
            not_free = (procedure not in self.__free)
            if not_free and condition:
                with condition:
                    self.__lock.release()
                    condition.wait()
                    self.__lock.acquire()
            elif not_free:
                self.__lock.wait()
        assert(not_free == False)
        assert(procedure in self.__free)
        return procedure

    def _procedure_enqueued(self, procedure):
        """Return information on the procedure if it is enqueued.
        Otherwise, raise an exception.
        """
        try:
            return self.__procedures[procedure]
        except KeyError:
            raise _errors.LockManagerError(
                "The procedure (%s) was never enqueued." % (procedure, )
            )



#
# Copyright (c) 2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
This file contains the command to check the data consistency in a replication
topology.
"""

from mysql.utilities.common.rpl_sync import RPLSynchronizer


def check_data_consistency(master_cnx_val, slaves_cnx_val, options,
                           data_to_include=None, data_to_exclude=None,
                           check_srv_versions=True):
    """
    Check the data consistency of a replication topology.

    This function creates a replication synchronizer checker and checks the
    data consistency between the given list of servers.

    master_cnx_val[in]      Dictionary with the connection values for the
                            master.
    slaves_cnx_val[in]      List of the dictionaries with the connection
                            values for each slave.
    options[in]             Dictionary of options (discover, verbosity,
                            rpl_timeout, checksum_timeout, interval).
    data_to_include[in]     Dictionary of data (set of tables) by database to
                            check.
    data_to_exclude[in]     Dictionary of data (set of tables) by database to
                            exclude from the check.
    check_srv_versions[in]  Flag indicating if the servers version check will
                            be performed. By default True, meaning that
                            differences between server versions will be
                            reported.

    Returns the number of issues found during the consistency check.
    """
    # Create replication synchronizer.
    rpl_sync = RPLSynchronizer(master_cnx_val, slaves_cnx_val, options)
    if check_srv_versions:
        # Check server versions and report differences.
        rpl_sync.check_server_versions()
    # Check GTID support, skipping slave with GTID disabled, and report
    # GTID executed differences between master and slaves.
    rpl_sync.check_gtid_sync()
    # Check data consistency and return the number of issues found.
    return rpl_sync.check_data_sync(options, data_to_include, data_to_exclude)


def check_server_versions(master_cnx_val, slaves_cnx_val, options):
    """
    Check the server versions of a replication topology.

    This method creates a replication synchronizer checker and compares the
    server versions of the given list of servers, reporting differences
    between them.

    master_cnx_val[in]  Dictionary with the connection values for the master.
    slaves_cnx_val[in]  List of the dictionaries with the connection values
                        for each slave.
    options[in]         Dictionary of options (discover, verbosity).
    """
    # Create replication synchronizer.
    rpl_sync = RPLSynchronizer(master_cnx_val, slaves_cnx_val, options)
    # Check server versions and report differences.
    rpl_sync.check_server_versions()



#
# Copyright (c) 2013,2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""Reading the configuration files and options.

This module implements support needed to read the configuration files
and set up the configuration used both by the Fabric nodes and by the
clients.

Reading the configuration goes through three steps:

1. Read the site-wide configuration file.

2. Read the command-line configuration file (if provided). Settings in
   this file will override any settings read from the site-wide
   configuration file.

3. Some options override settings in both the configuration files.

"""

import ConfigParser
import os
import re

# These are propagated to the importer
from ConfigParser import NoSectionError, NoOptionError

_VALUE_CRE = re.compile(
    r'(?P<section>\w+(?:\.\w+)*)\.(?P<name>\w+)=(?P<value>.*)')

def parse_param(option, _opt, value, parser):
    """Parser to parse a param option of the form x.y.z=some-value.

    This function is used as a callback to parse param values given on
    the command-line.
    """

    mobj = _VALUE_CRE.match(value)
    if mobj:
        field = getattr(parser.values, option.dest)
        if field is None:
            field = {}
            setattr(parser.values, option.dest, field)
        section = field.setdefault(mobj.group('section'), {})
        section[mobj.group('name')] = mobj.group('value')

class Config(ConfigParser.SafeConfigParser):
    """Fabric configuration file parser and configuration handler.

    This class manages the configuration of Fabric nodes and clients,
    including configuration file locations.

    Sample usage::

       from mysql.fabric.options import OptionParser
       from mysql.fabric.config import Config

       parser = OptionParser()
       ...
       options, args = parser.parse_args()
       config = Config(options.config_file, options.config_params)

    """

    def normalize_ssl_config(self, section):
        """Normalizes the SSL option in a section

        :param section: Section from which we read SSL configuration.
        """
        if not self.config_file:
            return
        default_path = os.path.dirname(self.config_file)
        try:
            for option in ('ssl_ca', 'ssl_key', 'ssl_cert'):
                value = self.get(section, option)
                if not value:
                    self.remove_option(section, option)
                elif not os.path.isabs(value):
                    self.set(section, option, os.path.join(default_path, value))
        except NoOptionError:
            # It's OK when SSL is missing
            pass

    def __init__(self, config_file, config_params=None):
        """Create the configuration parser, read the configuration
        files, and set up the configuration from the options.
        """

        ConfigParser.SafeConfigParser.__init__(self)

        if config_file is not None:
            self.readfp(open(config_file))
        self.config_file = config_file

        # Incorporate options into the configuration. These are read
        # from the mapping above and written into the configuration.
        if config_params is not None:
            for section, var_dict in config_params.items():
                if not self.has_section(section):
                    self.add_section(section)
                for key, val in var_dict.items():
                    self.set(section, key, val)

        self.normalize_ssl_config('protocol.xmlrpc')



#
# Copyright (c) 2013,2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

import optparse

import mysql.fabric.config as _config

class OptionParser(optparse.OptionParser):
    """Option with default options for all tools.

    This class is used as default option parser and specific commands
    can add their own options.

    There are three options that are provided:

    .. option:: --param <section>.<name>=<value> ...

       This option allow configuration parameters to be overridden by
       providing them on the command line.  The parameters are stored
       in the config_param attribute

    .. option:: --config <filename>

       Name of an extra configuration file to read, in addition to the
       site-wide configuration file. Options given in this file will
       override the options given in the site-wide file.

    Based on this, the options structure returned can hold the
    following attributes:

    .. attribute:: config_file

       File name for extra configuration file to read, provided by the
       :option:`--config` option. Defaults to ``fabric.cfg``.

    .. attribute:: config_param

       The configuration parameters provided with :option:`--param` as a
       dictionary of dictionaries, for example::

         {
            'protocol.xmlrpc': {
               'address': 'localhost:32274',
            },
            'logging': {
               'level': 'INFO',
            },
         }

    """
    def __init__(self, *args, **kwrds):
        optparse.OptionParser.__init__(self, *args, **kwrds)

        self.add_option(
            "--param",
            action="callback", callback=_config.parse_param,
            type="string", nargs=1, dest="config_params",
            help="Override a configuration parameter.")
        self.add_option(
            "--config",
            action="store", dest="config_file", default=None,
            metavar="FILE",
            help="Read configuration from FILE.")

    def print_error(self, msg):
        """Format and print an error message to the standard output.

        :param msg: Error message.
        """
        print "%s: %s\n" % (self.get_prog_name(), msg)

    def print_help(self):
        """Print help information to the standard output.
        """
        self.set_usage(
            "Usage: %prog [--param, --config] <grp> <cmd> [arg, ...]."
        )

        optparse.OptionParser.print_help(self)

        print "\n".join([
                "",
                "Basic commands:",
                "    help <grp> <cmd>  Show help for command",
                "    help commands     List all commands",
                "    help groups       List all groups",
        ])
